@@ -1,13 +1,13 @@
 #!/bin/sh
 
 # compile_cdrskin.sh  
-# Copyright 2005 - 2006 Thomas Schmitt, scdbackup@gmx.net, GPL
+# Copyright 2005 - 2008 Thomas Schmitt, scdbackup@gmx.net, GPL
 # to be executed within  ./libburn-*  resp ./cdrskin-*
 
-debug_opts=
+debug_opts="-O2"
 def_opts=
 largefile_opts="-D_FILE_OFFSET_BITS=64 -D_LARGEFILE_SOURCE=1"
-libvers="-DCdrskin_libburn_0_3_8"
+libvers="-DCdrskin_libburn_0_4_2"
 cleanup_src_or_obj="libburn/cleanup.o"
 libdax_msgs_o="libburn/libdax_msgs.o"
 libdax_audioxtr_o="libburn/libdax_audioxtr.o"
@@ -33,15 +33,15 @@ do
     libdax_audioxtr_o=
     libdax_msgs_o="libburn/message.o"
     cleanup_src_or_obj="-DCleanup_has_no_libburn_os_H cdrskin/cleanup.c"
-  elif test "$i" = "-libburn_0_3_8"
+  elif test "$i" = "-libburn_0_4_2"
   then
-    libvers="-DCdrskin_libburn_0_3_8"
+    libvers="-DCdrskin_libburn_0_4_2"
     libdax_audioxtr_o="libburn/libdax_audioxtr.o"
     libdax_msgs_o="libburn/libdax_msgs.o"
     cleanup_src_or_obj="libburn/cleanup.o"
   elif test "$i" = "-libburn_svn"
   then
-    libvers="-DCdrskin_libburn_0_3_9"
+    libvers="-DCdrskin_libburn_0_4_3"
     libdax_audioxtr_o="libburn/libdax_audioxtr.o"
     libdax_msgs_o="libburn/libdax_msgs.o"
     cleanup_src_or_obj="libburn/cleanup.o"
@@ -68,10 +68,7 @@ do
     do_strip=1
   elif test "$i" = "-g"
   then
-    debug_opts="$debug_opts -g"
-  elif test "$i" = "-O2"
-  then
-    debug_opts="$debug_opts -O2"
+    debug_opts="-g"
   elif test "$i" = "-help" -o "$i" = "--help" -o "$i" = "-h"
   then
     echo "cdrskin/compile_cdrskin.sh : to be executed within top level directory"
@@ -79,7 +76,7 @@ do
     echo "  -compile_cdrfifo  compile program cdrskin/cdrfifo."
     echo "  -compile_dewav    compile program test/dewav without libburn."
     echo "  -cvs_A60220       set macro to match libburn-CVS of 20 Feb 2006."
-    echo "  -libburn_0_3_8    set macro to match libburn-0.3.8."
+    echo "  -libburn_0_4_2    set macro to match libburn-0.4.2."
     echo "  -libburn_svn      set macro to match current libburn-SVN."
     echo "  -no_largefile     do not use 64 bit off_t (must match libburn)."
     echo "  -do_not_compile_cdrskin  omit compilation of cdrskin/cdrskin."
@@ -87,8 +84,7 @@ do
     echo "  -oldfashioned     use pre-0.2.2 libburn features only."
     echo "  -do_diet          produce capability reduced lean version."
     echo "  -do_strip         apply program strip to compiled programs."
-    echo "  -g                compile with cc option -g."
-    echo "  -O2               compile with cc option -O2."
+    echo "  -g                produce debuggable programm."
     echo "  -static           compile with cc option -static."
     exit 0
   elif test "$i" = "-static"
@@ -133,6 +129,7 @@ then
     \
     libburn/sg.o \
     libburn/write.o \
+    libburn/read.o \
     $libdax_audioxtr_o \
     $libdax_msgs_o \
     \
