@@ -16,7 +16,9 @@
 #include <string.h>
 #include <ctype.h>
 #include <pthread.h>
+#include <errno.h>
 #include "libburn.h"
+#include "init.h"
 #include "drive.h"
 #include "transport.h"
 #include "debug.h"
@@ -35,6 +37,9 @@
 /* A70225 : to learn about eventual Libburn_dvd_r_dl_multi_no_close_sessioN */
 #include "write.h"
 
+/* A70903 : for burn_scsi_setup_drive() */
+#include "spc.h"
+
 #include "libdax_msgs.h"
 extern struct libdax_msgs *libdax_messenger;
 
@@ -47,11 +52,39 @@ int burn_setup_drive(struct burn_drive *d, char *fname)
 {
 	d->devname = burn_strdup(fname);
 	memset(&d->params, 0, sizeof(struct params));
+	d->idata = NULL;
+	d->mdata = NULL;
+	d->toc_entry = NULL;
 	d->released = 1;
+	d->stdio_fd = -1;
 	d->status = BURN_DISC_UNREADY;
 	return 1;
 }
 
+
+/* ts A70903 */
+void burn_drive_free_subs(struct burn_drive *d)
+{
+	if (d->idata != NULL)
+		free((void *) d->idata);
+	d->idata = NULL;
+	if (d->mdata != NULL) {
+		burn_mdata_free_subs(d->mdata);
+		free((void *) d->mdata);
+	}
+	d->mdata = NULL;
+	if(d->toc_entry != NULL)
+		free((void *) d->toc_entry);
+	d->toc_entry = NULL;
+	if (d->devname != NULL)
+		free(d->devname);
+	d->devname = NULL;
+	if (d->stdio_fd >= 0)
+		close (d->stdio_fd);
+	d->stdio_fd = -1;
+}
+
+
 /* ts A60904 : ticket 62, contribution by elmom */
 /* splitting former burn_drive_free() (which freed all, into two calls) */
 void burn_drive_free(struct burn_drive *d)
@@ -59,14 +92,12 @@ void burn_drive_free(struct burn_drive *d)
 	if (d->global_index == -1)
 		return;
 	/* ts A60822 : close open fds before forgetting them */
-	if (burn_drive_is_open(d))
-		d->release(d);
-	free((void *) d->idata);
-	burn_mdata_free_subs(d->mdata);
-	free((void *) d->mdata);
-	if(d->toc_entry != NULL)
-		free((void *) d->toc_entry);
-	free(d->devname);
+	if (d->drive_role == 1)
+		if (burn_drive_is_open(d)) {
+			d->unlock(d);
+			d->release(d);
+		}
+	burn_drive_free_subs(d);
 	d->global_index = -1;
 }
 
@@ -84,6 +115,8 @@ void burn_drive_free_all(void)
 /* ts A60822 */
 int burn_drive_is_open(struct burn_drive *d)
 {
+	if (d->drive_role != 1)
+		return (d->stdio_fd >= 0);
 	/* ts A61021 : moved decision to sg.c */
 	return d->drive_is_open(d);
 }
@@ -110,6 +143,7 @@ int burn_drive_is_released(struct burn_drive *d)
                   -1 = drive is closed (i.e. released explicitely)
                    0 = drive is open, not grabbed (after scan, before 1st grab)
                    1 = drive is grabbed but BURN_DRIVE_IDLE
+                   2 = drive is grabbed, synchronous read/write interrupted
                   10 = drive is grabbing (BURN_DRIVE_GRABBING)
                  100 = drive is busy in cancelable state
                 1000 = drive is in non-cancelable state 
@@ -127,8 +161,21 @@ int burn_drive_is_occupied(struct burn_drive *d)
 		return 0;
 	if(d->busy == BURN_DRIVE_IDLE)
 		return 1;
-	if(d->busy == BURN_DRIVE_READING || d->busy == BURN_DRIVE_WRITING)
+	if(d->busy == BURN_DRIVE_READING_SYNC ||
+		 d->busy == BURN_DRIVE_WRITING_SYNC)
+		return 2;
+	if(d->busy == BURN_DRIVE_WRITING) {
+
+		/* ts A70928 */
+		/* >>> how do i learn whether the writer thread is still
+			 alive ? */;
+			/* >>> what to do if writer is dead ? 
+				At least sync disc ?*/;
 		return 50;
+	}
+	if(d->busy == BURN_DRIVE_READING) {
+		return 50;
+	}
 	return 1000;
 }
 
@@ -154,7 +201,9 @@ int burn_drive_inquire_media(struct burn_drive *d)
 
 	/* ts A61020 : d->status was set to BURN_DISC_BLANK as pure guess */
 
-	if (d->mdata->cdr_write || d->mdata->cdrw_write ||
+        /* ts A71128 : run read_disc_info() for any recognizeable profile */
+	if (d->current_profile > 0 || 
+	    d->mdata->cdr_write || d->mdata->cdrw_write ||
 	    d->mdata->dvdr_write || d->mdata->dvdram_write) {
 
 #define Libburn_knows_correct_state_after_loaD 1
@@ -224,6 +273,19 @@ int burn_drive_grab(struct burn_drive *d, int le)
 		burn_print(1, "can't grab - already grabbed\n");
 		return 0;
 	}
+	if(d->drive_role != 1) {
+		d->released = 0;
+		if(d->drive_role == 2 || d->drive_role == 3) {
+			d->status = BURN_DISC_BLANK;
+			d->current_profile = 0xffff;
+		} else {
+			d->status = BURN_DISC_EMPTY;
+			d->current_profile = 0;
+		}
+		d->busy = BURN_DRIVE_IDLE;
+		return 1;
+	}
+
 	d->status = BURN_DISC_UNREADY;
 	errcode = d->grab(d);
 
@@ -252,6 +314,10 @@ int burn_drive_grab(struct burn_drive *d, int le)
 	return ret;
 }
 
+
+/* ts A71015 */
+#define Libburn_ticket_62_re_register_is_possiblE 1
+
 struct burn_drive *burn_drive_register(struct burn_drive *d)
 {
 #ifdef Libburn_ticket_62_re_register_is_possiblE
@@ -268,6 +334,8 @@ struct burn_drive *burn_drive_register(struct burn_drive *d)
 	d->rlba = 0;
 	d->cancel = 0;
 	d->busy = BURN_DRIVE_IDLE;
+	d->thread_pid = 0;
+	d->thread_pid_valid = 0;
 	d->toc_entries = 0;
 	d->toc_entry = NULL;
 	d->disc = NULL;
@@ -276,6 +344,7 @@ struct burn_drive *burn_drive_register(struct burn_drive *d)
 #ifdef Libburn_ticket_62_re_register_is_possiblE
 	/* ts A60904 : ticket 62, contribution by elmom */
 	/* Not yet accepted because no use case seen yet */
+        /* ts A71015 : xorriso dialog imposes a use case now */
 
 	/* This is supposed to find an already freed drive struct among
 	   all the the ones that have been used before */
@@ -321,12 +390,14 @@ struct burn_drive *burn_drive_finish_enum(struct burn_drive *d)
 	struct burn_drive *t;
 	/* ts A60821
    	<<< debug: for tracing calls which might use open drive fds */
-	int mmc_function_spy(char * text);
+	int mmc_function_spy(struct burn_drive *d, char * text);
+
+	d->drive_role = 1; /* MMC drive */
 
 	t = burn_drive_register(d);
 
 	/* ts A60821 */
-	mmc_function_spy("enumerate_common : -------- doing grab");
+	mmc_function_spy(NULL, "enumerate_common : -------- doing grab");
 
 	/* try to get the drive info */
 	if (t->grab(t)) {
@@ -341,7 +412,7 @@ struct burn_drive *burn_drive_finish_enum(struct burn_drive *d)
 	}
 
 	/* ts A60821 */
-	mmc_function_spy("enumerate_common : ----- would release ");
+	mmc_function_spy(NULL, "enumerate_common : ----- would release ");
 
 	return t;
 }
@@ -372,7 +443,10 @@ int burn_drive_mark_unready(struct burn_drive *d)
 }
 
 
-void burn_drive_release(struct burn_drive *d, int le)
+/* ts A70918 : outsourced from burn_drive_release() and enhanced */
+/** @param flag bit0-2 = mode : 0=unlock , 1=unlock+eject , 2=leave locked
+*/
+int burn_drive_release_fl(struct burn_drive *d, int flag)
 {
 	if (d->released) {
 		/* ts A61007 */
@@ -381,7 +455,7 @@ void burn_drive_release(struct burn_drive *d, int le)
 				d->global_index, 0x00020105,
 				LIBDAX_MSGS_SEV_SORRY, LIBDAX_MSGS_PRIO_HIGH,
 				"Drive is already released", 0, 0);
-		return;
+		return 0;
 	}
 
 	/* ts A61007 */
@@ -392,23 +466,46 @@ void burn_drive_release(struct burn_drive *d, int le)
 				d->global_index, 0x00020106,
 				LIBDAX_MSGS_SEV_SORRY, LIBDAX_MSGS_PRIO_HIGH,
 				"Drive is busy on attempt to close", 0, 0);
-		return;
+		return 0;
 	}
 
-	d->unlock(d);
-	if (le)
-		d->eject(d);
-	d->release(d);
+	if (d->drive_role == 1) {
+		if (d->needs_sync_cache)
+			d->sync_cache(d);
+		if ((flag & 7) != 2)
+			d->unlock(d);
+		if ((flag & 7) == 1)
+			d->eject(d);
+		d->release(d);
+	}
+
+	d->needs_sync_cache = 0; /* just to be sure */
 	d->released = 1;
 
 	/* ts A61125 : outsourced model aspects */
 	burn_drive_mark_unready(d);
+	return 1;
 }
 
 
+/* API */
+void burn_drive_release(struct burn_drive *d, int le)
+{
+	burn_drive_release_fl(d, !!le);
+}
+
+
+/* ts A70918 */
+/* API */
+int burn_drive_leave_locked(struct burn_drive *d, int flag)
+{
+	return burn_drive_release_fl(d, 2);
+}
+
 
 /* ts A61007 : former void burn_wait_all() */
-int burn_drives_are_clear(void)
+/* @param flag  bit0= demand freed drives (else released drives) */
+int burn_drives_are_clear(int flag)
 {
 	int i;
 
@@ -416,7 +513,7 @@ int burn_drives_are_clear(void)
 		/* ts A60904 : ticket 62, contribution by elmom */
 		if (drive_array[i].global_index == -1)
 	continue;
-		if (drive_array[i].released)
+		if (drive_array[i].released && !(flag & 1))
 	continue;
 		return 0;
 	}
@@ -481,7 +578,8 @@ void burn_disc_erase_sync(struct burn_drive *d, int fast)
 
 	/* ts A61125 : update media state records */
 	burn_drive_mark_unready(d);
-	burn_drive_inquire_media(d);
+	if (d->drive_role == 1)
+		burn_drive_inquire_media(d);
 	d->busy = BURN_DRIVE_IDLE;
 }
 
@@ -509,6 +607,7 @@ void burn_disc_format_sync(struct burn_drive *d, off_t size, int flag)
 	stages = 1 + ((flag & 1) && size > 1024 * 1024);
 	d->cancel = 0;
 	d->busy = BURN_DRIVE_FORMATTING;
+
 	ret = d->format_unit(d, size, flag & 0xff96); /* forward bits */
 	if (ret <= 0)
 		d->cancel = 1;
@@ -589,6 +688,8 @@ int burn_disc_get_formats(struct burn_drive *d, int *status, off_t *size,
 	*size = 0;
 	*bl_sas = 0;
 	*num_formats = 0;
+	if (d->drive_role != 1)
+		return 0;
 	ret = d->read_format_capacities(d, 0x00);
 	if (ret <= 0)
 		return 0;
@@ -639,7 +740,20 @@ int burn_disc_erasable(struct burn_drive *d)
 enum burn_drive_status burn_drive_get_status(struct burn_drive *d,
 					     struct burn_progress *p)
 {
-	if (p) {
+	/* ts A70928 : inform control thread of signal in sub-threads */
+	if (burn_global_abort_level > 0)
+		burn_global_abort_level++;
+	if (burn_global_abort_level > 5) {
+		if (burn_global_signal_handler == NULL)
+			kill(getpid(), burn_global_abort_signum);
+		else
+			(*burn_global_signal_handler)
+				(burn_global_signal_handle,
+				 burn_global_abort_signum, 0);
+		burn_global_abort_level = -1;
+	}
+
+	if (p != NULL) {
 		memcpy(p, &(d->progress), sizeof(struct burn_progress));
 		/* TODO: add mutex */
 	}
@@ -733,80 +847,101 @@ static int drive_getcaps(struct burn_drive *d, struct burn_drive_info *out)
 	return 1;
 }
 
+
+/* ts A70907 : added parameter flag */
+/* @param flag bit0= reset global drive list */
 int burn_drive_scan_sync(struct burn_drive_info *drives[],
-			 unsigned int *n_drives)
+			 unsigned int *n_drives, int flag)
 {
-	/* state vars for the scan process */
-	/* ts A60904 : did set some default values to feel comfortable */
-	static int scanning = 0, scanned = 0, found = 0;
-	static unsigned num_scanned = 0, count = 0;
-	unsigned int i;
+	/* ts A70907 :
+	   There seems to have been a misunderstanding about the role of
+	   burn_drive_scan_sync(). It needs no static state because it
+	   is only started once during an asynchronous scan operation.
+	   Its starter, burn_drive_scan(), is the one which ends immediately
+	   and gets called repeatedly. It acts on start of scanning by
+	   calling burn_drive_scan_sync(), returns idle while scanning is
+	   not done and finally removes the worker object which represented
+	   burn_drive_scan_sync().
+	   The scanning itself is not parallel but enumerates sequentially
+	   drive by drive (within scsi_enumerate_drives()).
+
+	   I will use "scanned" for marking drives found by previous runs.
+           It will not be static any more.
+	*/
+	/* ts A71015 : this makes only trouble : static int scanning = 0; */
+	/* ts A70907 :
+	   These variables are too small anyway. We got up to 255 drives.
+	   static int scanned = 0, found = 0;
+	   Variable "found" was only set but never read.
+	*/
+	unsigned char scanned[32];
+	unsigned count = 0;
+	int i;
 
 	/* ts A61007 : moved up to burn_drive_scan() */
 	/* a ssert(burn_running); */
 
-	if (!scanning) {
-		scanning = 1;
 
-		/* ts A61007 : test moved up to burn_drive_scan()
+	/* ts A61007 : test moved up to burn_drive_scan()
 		               burn_wait_all() is obsoleted */
 #if 0
-		/* make sure the drives aren't in use */
-		burn_wait_all();	/* make sure the queue cleans up
-					   before checking for the released
-					   state */
+	/* make sure the drives aren't in use */
+	burn_wait_all();	/* make sure the queue cleans up
+				   before checking for the released
+				   state */
 #endif /* 0 */
 
-		/* refresh the lib's drives */
+	*n_drives = 0;
 
-		/* ts A61115 : formerly sg_enumerate(); ata_enumerate(); */
-		scsi_enumerate_drives();
-
-		count = burn_drive_count();
-		if (count)
-			*drives =
-				malloc(sizeof(struct burn_drive_info) * count);
-		else
-			*drives = NULL;
-		*n_drives = scanned = found = num_scanned = 0;
+	/* ts A70907 : wether to scan from scratch or to extend */
+	for (i = 0; i < sizeof(scanned); i++)
+		scanned[i] = 0;
+	if (flag & 1) {
+		burn_drive_free_all();
+	} else {
+		for (i = 0; i <= drivetop; i++) 
+			if (drive_array[i].global_index >= 0)
+				scanned[i / 8] |=  (1 << (i % 8));
 	}
 
+	/* refresh the lib's drives */
+
+	/* ts A61115 : formerly sg_enumerate(); ata_enumerate(); */
+	scsi_enumerate_drives();
+
+	count = burn_drive_count();
+	if (count) {
+		/* ts A70907 :
+		   Extra array element marks end of array. */
+		*drives = calloc(count + 1,
+				sizeof(struct burn_drive_info));
+		if (*drives == NULL) {
+	       		libdax_msgs_submit(libdax_messenger, -1, 0x00000003,
+	               			LIBDAX_MSGS_SEV_FATAL,
+					LIBDAX_MSGS_PRIO_HIGH,
+					"Out of virtual memory", 0, 0);
+			return -1;
+		} else
+			for (i = 0; i <= count; i++) /* invalidate */
+				(*drives)[i].drive = NULL;
+	} else
+		*drives = NULL;
+
 	for (i = 0; i < count; ++i) {
-		if (scanned & (1 << i))
-			continue;	/* already scanned the device */
+		if (scanned[i / 8] & (1 << (i % 8)))
+	continue;		/* device already scanned by previous run */
+		if (drive_array[i].global_index < 0)
+	continue;		/* invalid device */
 
 		while (!drive_getcaps(&drive_array[i],
-				      &(*drives)[num_scanned])) {
+				      &(*drives)[*n_drives])) {
 			sleep(1);
 		}
-		scanned |= 1 << i;
-		found |= 1 << i;
-		num_scanned++;
 		(*n_drives)++;
+		scanned[i / 8] |= 1 << (i % 8);
 	}
 
-	if (num_scanned == count) {
-		/* done scanning */
-		scanning = 0;
-		return 1;
-	}
-	return 0;
-}
-
-
-void burn_drive_info_free(struct burn_drive_info drive_infos[])
-{
-/* ts A60904 : ticket 62, contribution by elmom */
-/* clarifying the meaning and the identity of the victim */
-
-	/* ts A60904 : This looks a bit weird.
-	   burn_drive_info is not the manager of burn_drive but only its
-	   spokesperson. To my knowlege drive_infos from burn_drive_scan()
-	   are not memorized globally. */
-	if(drive_infos != NULL)
-		free((void *) drive_infos);
-
-	burn_drive_free_all();
+	return(1);
 }
 
 /* ts A61001 : internal call */
@@ -839,9 +974,66 @@ int burn_drive_forget(struct burn_drive *d, int force)
 /* API call */
 int burn_drive_info_forget(struct burn_drive_info *info, int force)
 {
-  return burn_drive_forget(info->drive, force);
+	return burn_drive_forget(info->drive, force);
 }
 
+
+void burn_drive_info_free(struct burn_drive_info drive_infos[])
+{
+	int i;
+
+/* ts A60904 : ticket 62, contribution by elmom */
+/* clarifying the meaning and the identity of the victim */
+
+	if(drive_infos == NULL)
+		return;
+
+#ifndef Libburn_free_all_drives_on_infO
+
+#ifdef Not_yeT
+	int new_drivetop;
+
+	/* ts A71015: compute reduced drivetop counter */
+	new_drivetop = drivetop;
+	for (i = 0; drive_infos[i].drive != NULL; i++)
+		if (drive_infos[i].global_index == new_drivetop
+		    && new_drivetop >= 0) {
+			new_drivetop--;
+			i = 0;
+		}
+#endif /* Not_yeT */
+
+	/* ts A70907 : Solution for wrong behavior below */
+	for (i = 0; drive_infos[i].drive != NULL; i++)
+		burn_drive_free(drive_infos[i].drive);
+
+#ifdef Not_yeT
+	drivetop = new_drivetop;
+#endif /* Not_yeT */
+
+#endif /* ! Libburn_free_all_drives_on_infO */
+
+	/* ts A60904 : This looks a bit weird. [ts A70907 : not any more]
+	   burn_drive_info is not the manager of burn_drive but only its
+	   spokesperson. To my knowlege drive_infos from burn_drive_scan()
+	   are not memorized globally. */
+	free((void *) drive_infos);
+
+#ifdef Libburn_free_all_drives_on_infO
+	/* ts A70903 : THIS IS WRONG !      (disabled now)
+	   It endangers multi drive usage.
+	   This call is not entitled to delete all drives, only the
+	   ones of the array which it recieves a parmeter.
+
+	   Problem:  It was unclear how many items are listed in drive_infos
+           Solution: Added a end marker element to any burn_drive_info array 
+                     The mark can be recognized by having drive == NULL
+	*/
+	burn_drive_free_all();
+#endif
+}
+
+
 struct burn_disc *burn_drive_get_disc(struct burn_drive *d)
 {
 	/* ts A61022: SIGSEGV on calling this function with blank media */
@@ -854,6 +1046,9 @@ struct burn_disc *burn_drive_get_disc(struct burn_drive *d)
 
 void burn_drive_set_speed(struct burn_drive *d, int r, int w)
 {
+	d->nominal_write_speed = w;
+	if(d->drive_role != 1)
+		return;
 	d->set_speed(d, r, w);
 }
 
@@ -900,17 +1095,23 @@ void burn_sectors_to_msf(int sectors, int *m, int *s, int *f)
 
 int burn_drive_get_read_speed(struct burn_drive *d)
 {
+	if(!d->mdata->valid)
+		return 0;
 	return d->mdata->max_read_speed;
 }
 
 int burn_drive_get_write_speed(struct burn_drive *d)
 {
+	if(!d->mdata->valid)
+		return 0;
 	return d->mdata->max_write_speed;
 }
 
 /* ts A61021 : New API function */
 int burn_drive_get_min_write_speed(struct burn_drive *d)
 {
+	if(!d->mdata->valid)
+		return 0;
 	return d->mdata->min_write_speed;
 }
 
@@ -957,6 +1158,139 @@ int burn_drive_is_banned(char *device_address)
 	return 1;
 }
 
+
+/* ts A70924 */
+int burn_drive__fd_from_special_adr(char *adr)
+{
+	int fd = -1, i;
+
+	if (strcmp(adr, "-") == 0)
+		fd = 1;
+	if(strncmp(adr, "/dev/fd/", 8) == 0) {
+		for (i = 8; adr[i]; i++)
+			if (!isdigit(adr[i]))
+		break;
+		if (i> 8 && adr[i] == 0)
+			fd = atoi(adr + 8);
+	}
+	return fd;
+}
+
+
+/* ts A70903 : Implements adquiration of pseudo drives */
+int burn_drive_grab_dummy(struct burn_drive_info *drive_infos[], char *fname)
+{
+	int ret = -1, fd = -1, role = 0;
+	/* divided by 512 it needs to fit into a signed long integer */
+	off_t size = ((off_t) (512 * 1024 * 1024 - 1) * (off_t) 2048);
+	struct burn_drive *d= NULL, *regd_d;
+	struct stat stbuf;
+
+	static int allow_role_3 = 1;
+
+	if (fname[0] != 0) {
+		memset(&stbuf, 0, sizeof(stbuf));
+		fd = burn_drive__fd_from_special_adr(fname);
+		if (fd >= 0)
+			ret = fstat(fd, &stbuf);
+		else
+			ret = stat(fname, &stbuf);
+		if (ret == -1 || S_ISBLK(stbuf.st_mode) ||
+				S_ISREG(stbuf.st_mode)) {
+			ret = burn_os_stdio_capacity(fname, &size);
+			if (ret == -1) {
+				libdax_msgs_submit(libdax_messenger, -1,
+				 0x00020009,
+				 LIBDAX_MSGS_SEV_SORRY, LIBDAX_MSGS_PRIO_HIGH,
+				 "Neither stdio-path nor its directory exist",
+				 0, 0);
+				return 0;
+			} else if (ret == -2) {
+				libdax_msgs_submit(libdax_messenger, -1,
+				0x00020005,
+				LIBDAX_MSGS_SEV_SORRY, LIBDAX_MSGS_PRIO_HIGH,
+				"Failed to open device (a pseudo-drive)",
+				errno, 0);
+				return 0;
+			}
+			if (fname[0] != 0)
+				role = 2;
+			else
+				role = 0;
+		} else {
+			if(S_ISDIR(stbuf.st_mode) || !allow_role_3) {
+				libdax_msgs_submit(libdax_messenger, -1,
+				 0x00020149,
+				 LIBDAX_MSGS_SEV_SORRY, LIBDAX_MSGS_PRIO_HIGH,
+				 "Unsuitable filetype for pseudo-drive", 0, 0);
+				return 0;
+			}
+			role = 3;
+		}
+	}
+	d= (struct burn_drive *) calloc(1, sizeof(struct burn_drive));
+	if (d == NULL)
+		return 0;
+	burn_setup_drive(d, fname);
+	d->status = BURN_DISC_EMPTY;
+
+	d->drive_role = role;
+	ret = burn_scsi_setup_drive(d, -1, -1, -1, -1, -1, 0);
+	if (ret <= 0)
+		goto ex;
+	regd_d = burn_drive_register(d);
+	if (regd_d == NULL) {
+		ret = -1;
+		goto ex;
+	}
+	free((char *) d); /* all sub pointers have been copied to *regd_d */
+	d = regd_d;
+	if (d->drive_role == 2 || d->drive_role == 3) {
+		d->status = BURN_DISC_BLANK;
+		d->current_profile = 0xffff; /* MMC for non-compliant drive */
+		strcpy(d->current_profile_text,"stdio file");
+		d->current_is_cd_profile = 0;
+		d->current_is_supported_profile = 1;
+		d->block_types[BURN_WRITE_TAO] = BURN_BLOCK_MODE1;
+		d->block_types[BURN_WRITE_SAO] = BURN_BLOCK_SAO;
+		d->media_capacity_remaining = size;
+
+		/* >>> ? open file for a test ? (>>> beware of "-" = stdin) */;
+
+	} else
+		d->current_profile = 0; /* Drives return this if empty */
+
+	*drive_infos = calloc(2, sizeof(struct burn_drive_info));
+	if (*drive_infos == NULL)
+		goto ex;
+	(*drive_infos)[0].drive = d;
+	(*drive_infos)[1].drive = NULL; /* End-Of-List mark */
+	(*drive_infos)[0].tao_block_types = d->block_types[BURN_WRITE_TAO];
+	(*drive_infos)[0].sao_block_types = d->block_types[BURN_WRITE_SAO];
+	if (d->drive_role == 2) {
+		strcpy((*drive_infos)[0].vendor,"YOYODYNE");
+		strcpy((*drive_infos)[0].product,"WARP DRIVE");
+		strcpy((*drive_infos)[0].revision,"FX01");
+	} else if (d->drive_role == 3) {
+		strcpy((*drive_infos)[0].vendor,"YOYODYNE");
+		strcpy((*drive_infos)[0].product,"BLACKHOLE");
+		strcpy((*drive_infos)[0].revision,"FX02");
+	} else {
+		strcpy((*drive_infos)[0].vendor,"FERENGI");
+		strcpy((*drive_infos)[0].product,"VAPORWARE");
+		strcpy((*drive_infos)[0].revision,"0000");
+	}
+	d->released = 0;
+	ret = 1;
+ex:;
+	if (ret <= 0 && d != NULL) {
+		burn_drive_free_subs(d);
+		free((char *) d);
+	}
+	return ret;
+}
+
+
 /* ts A60823 */
 /** Aquire a drive with known persistent address. 
 */
@@ -964,7 +1298,26 @@ int burn_drive_scan_and_grab(struct burn_drive_info *drive_infos[], char* adr,
 			     int load)
 {
 	unsigned int n_drives;
-	int ret;
+	int ret, i;
+
+	/* check wether drive adress is already registered */
+	for (i = 0; i <= drivetop; i++)
+		if (drive_array[i].global_index >= 0)
+			if (strcmp(drive_array[i].devname, adr) == 0)
+	break;
+	if (i <= drivetop) {
+		libdax_msgs_submit(libdax_messenger, i,
+				0x0002014b,
+				LIBDAX_MSGS_SEV_SORRY, LIBDAX_MSGS_PRIO_HIGH,
+				"Drive is already registered resp. scanned",
+				0, 0);
+		return -1;
+	}
+
+	if (strncmp(adr, "stdio:", 6) == 0) {
+		ret = burn_drive_grab_dummy(drive_infos, adr + 6);
+		return ret;
+	}
 
 	burn_drive_clear_whitelist();
 	burn_drive_add_whitelist(adr);
@@ -972,18 +1325,20 @@ int burn_drive_scan_and_grab(struct burn_drive_info *drive_infos[], char* adr,
 	fprintf(stderr,"libburn: experimental: burn_drive_scan_and_grab(%s)\n",
 		adr);
 */
-	while (1) {
-		ret = burn_drive_scan(drive_infos, &n_drives);
-		if (ret < 0)
-			return -1;
-		if (ret > 0)
-	break;
-		usleep(1002);
-	}
+
+	/* ts A70907 : now calling synchronously rather than looping */
+	ret = burn_drive_scan_sync(drive_infos, &n_drives, 0);
+	if (ret < 0)
+		return -1;
+
 	if (n_drives <= 0)
 		return 0;
 /*
-	fprintf(stderr, "libburn: experimental: n_drives == %d\n",n_drives);
+	fprintf(stderr, "libburn: experimental: n_drives %d , drivetop %d\n",
+		n_drives, drivetop);
+	if (n_drives > 0)
+		fprintf(stderr, "libburn: experimental: global_index %d\n",
+			drive_infos[0]->drive->global_index);
 */
 
 	ret = burn_drive_grab(drive_infos[0]->drive, load);
@@ -1015,9 +1370,9 @@ int burn_drive_adr_debug_msg(char *fmt, char *arg)
 	return ret;
 }
 
-/* ts A60923 */
+/* ts A60923 */ /* ts A70906 : promoted to API */
 /** Inquire the persistent address of the given drive. */
-int burn_drive_raw_get_adr(struct burn_drive *d, char adr[])
+int burn_drive_d_get_adr(struct burn_drive *d, char adr[])
 {
 	if (strlen(d->devname) >= BURN_DRIVE_ADR_LEN) {
 		libdax_msgs_submit(libdax_messenger, d->global_index,
@@ -1030,17 +1385,19 @@ int burn_drive_raw_get_adr(struct burn_drive *d, char adr[])
 	return 1;
 }
 
-/* ts A60823 - A60923 */
+/* ts A60823 - A60923 */ /* A70906 : Now legacy API call */
 /** Inquire the persistent address of the given drive. */
 int burn_drive_get_adr(struct burn_drive_info *drive_info, char adr[])
 {
 	int ret;
 
-	ret = burn_drive_raw_get_adr(drive_info->drive, adr);
+	ret = burn_drive_d_get_adr(drive_info->drive, adr);
 	return ret;
 }
 
 
+
+
 /* ts A60922 ticket 33 */
 /** Evaluate wether the given address would be enumerated by libburn */
 int burn_drive_is_enumerable_adr(char *adr)
@@ -1141,7 +1498,7 @@ int burn_drive_obtain_scsi_adr(char *path,
 	for (i = 0; i < drivetop + 1; i++) {
 		if (drive_array[i].global_index < 0)
 	continue;
-		ret = burn_drive_raw_get_adr(&(drive_array[i]),adr);
+		ret = burn_drive_d_get_adr(&(drive_array[i]),adr);
 		if (ret < 0)
 			return -1;
 		if (ret == 0)
@@ -1246,7 +1603,8 @@ int burn_drive_convert_fs_adr_sub(char *path, char adr[], int *rec_count)
 	struct stat stbuf;
 
 	burn_drive_adr_debug_msg("burn_drive_convert_fs_adr( %s )", path);
-	if(burn_drive_is_enumerable_adr(path)) {
+	if (strncmp(path, "stdio:", 6) == 0 ||
+	    burn_drive_is_enumerable_adr(path)) {
 		if(strlen(path) >= BURN_DRIVE_ADR_LEN)
 			return -1;
 		burn_drive_adr_debug_msg(
@@ -1327,6 +1685,12 @@ int burn_abort(int patience,
 
 	current_time = start_time = pacifier_time = time(0);
 	end_time = start_time + patience;
+
+	/* >>> ts A71002 : are there any threads at work ?
+		If not, then one can force abort because the drives will not
+		change status on their own.
+	 */
+
 	while(current_time-end_time < patience) {
 		still_not_done = 0;
 
@@ -1379,6 +1743,8 @@ int burn_drive_get_start_end_lba(struct burn_drive *d,
 /* ts A61020 API function */
 int burn_disc_pretend_blank(struct burn_drive *d)
 {
+	if (d->drive_role == 0)
+		return 0;
 	if (d->status != BURN_DISC_UNREADY && 
 	    d->status != BURN_DISC_UNSUITABLE)
 		return 0;
@@ -1389,6 +1755,8 @@ int burn_disc_pretend_blank(struct burn_drive *d)
 /* ts A61106 API function */
 int burn_disc_pretend_full(struct burn_drive *d)
 {
+	if (d->drive_role == 0)
+		return 0;
 	if (d->status != BURN_DISC_UNREADY && 
 	    d->status != BURN_DISC_UNSUITABLE)
 		return 0;
@@ -1407,6 +1775,8 @@ int burn_disc_read_atip(struct burn_drive *d)
 				0, 0);
 		return -1;
 	}
+	if(d->drive_role != 1)
+		return 0;
 	if (d->current_profile == -1 || d->current_is_cd_profile) {
 		d->read_atip(d);
 		/* >>> some control of success would be nice :) */
@@ -1439,6 +1809,9 @@ int burn_disc_track_lba_nwa(struct burn_drive *d, struct burn_write_opts *o,
 			0, 0);
 		return -1;
 	}
+	*lba = *nwa = 0;
+	if (d->drive_role != 1)
+		return 0;
 	if (o != NULL)
 		d->send_write_parameters(d, o);
 	ret = d->get_nwa(d, trackno, lba, nwa);
@@ -1467,6 +1840,9 @@ int burn_disc_get_msc1(struct burn_drive *d, int *start)
 			0, 0);
 		return -1;
 	}
+	*start = 0;
+	if (d->drive_role != 1)
+		return 0;
 	ret = d->read_multi_session_c1(d, &trackno, start);
 	return ret;
 }
@@ -1479,13 +1855,20 @@ off_t burn_disc_available_space(struct burn_drive *d,
 	int lba, nwa;
 
 	if (burn_drive_is_released(d))
-		goto ex;
+		return 0;
 	if (d->busy != BURN_DRIVE_IDLE)
-		goto ex;
-	if (o != NULL)
-		d->send_write_parameters(d, o);
-	d->get_nwa(d, -1, &lba, &nwa);
-ex:;
+		return 0;
+	if (d->drive_role == 0)
+		return 0;
+	if (d->drive_role != 1) {
+		if (d->media_capacity_remaining <= 0)
+			d->media_capacity_remaining = 
+			((off_t) (512 * 1024 * 1024 - 1) * (off_t) 2048);
+	} else {
+		if (o != NULL)
+			d->send_write_parameters(d, o);
+		d->get_nwa(d, -1, &lba, &nwa);
+	}
 	if (o != NULL) {
 		if (o->start_byte > 0) {
 			if (o->start_byte > d->media_capacity_remaining)
@@ -1590,6 +1973,8 @@ int burn_speed_descriptor_copy(struct burn_speed_descriptor *from,
 /* ts A61226 : free dynamically allocated sub data of struct scsi_mode_data */
 int burn_mdata_free_subs(struct scsi_mode_data *m)
 {
+	if(!m->valid)
+		return 0;
 	burn_speed_descriptor_destroy(&(m->speed_descriptors), 1);
 	return 1;
 }
@@ -1603,6 +1988,8 @@ int burn_drive_get_speedlist(struct burn_drive *d,
 	struct burn_speed_descriptor *sd, *csd = NULL;
 
 	(*speed_list) = NULL;
+	if(!d->mdata->valid)
+		return 0;
 	for (sd = d->mdata->speed_descriptors; sd != NULL; sd = sd->next) {
 		ret = burn_speed_descriptor_new(&csd, NULL, csd, 0);
 		if (ret <= 0)
@@ -1626,6 +2013,8 @@ int burn_drive_get_best_speed(struct burn_drive *d, int speed_goal,
 	if (speed_goal < 0)
 		best_speed = 2000000000;
 	*best_descr = NULL;
+	if(!d->mdata->valid)
+		return 0;
 	for (sd = d->mdata->speed_descriptors; sd != NULL; sd = sd->next) {
 		if (flag & 1)
 			speed = sd->read_speed;
@@ -1699,7 +2088,27 @@ int burn_disc_get_multi_caps(struct burn_drive *d, enum burn_write_types wt,
 	o->current_is_cd_profile = d->current_is_cd_profile;
         o->might_simulate = 0;
 	
-	if (s != BURN_DISC_BLANK && s != BURN_DISC_APPENDABLE) {
+	if (d->drive_role == 0)
+		return 0;
+	if (d->drive_role == 2) {
+		/* stdio file drive : random access read-write */
+		o->start_adr = 1;
+		size = d->media_capacity_remaining;
+		burn_os_stdio_capacity(d->devname, &size);
+		d->media_capacity_remaining = size;
+		o->start_range_high = size;
+		o->start_alignment = 2048; /* imposting a drive, not a file */
+		o->might_do_sao = 4;
+		o->might_do_tao = 2;
+		o->advised_write_mode = BURN_WRITE_TAO;
+        	o->might_simulate = 1;
+	} else if (d->drive_role != 1) {
+		/* stdio file drive : sequential access write-only */
+		o->might_do_sao = 4;
+		o->might_do_tao = 2;
+		o->advised_write_mode = BURN_WRITE_TAO;
+        	o->might_simulate = 1;
+	} else if (s != BURN_DISC_BLANK && s != BURN_DISC_APPENDABLE) {
 		return 0;
 	} else if (s == BURN_DISC_APPENDABLE &&
 		 (wt == BURN_WRITE_SAO || wt == BURN_WRITE_RAW)) {
@@ -1876,3 +2285,162 @@ int burn_disc_get_write_mode_demands(struct burn_disc *disc,
 	}
 	return (disc->sessions > 0);
 }
+
+
+/* ts A70903 : API */
+int burn_drive_get_drive_role(struct burn_drive *d)
+{
+	return d->drive_role;
+}
+
+
+/* ts A70923
+   Hands out pointers *dpt to directory path and *npt to basename.
+   Caution: the last '/' in adr gets replaced by a 0.
+*/
+static int burn__split_path(char *adr, char **dpt, char **npt)
+{
+	*dpt = adr;
+	*npt = strrchr(*dpt, '/');
+	if (*npt == NULL) {
+		*npt = *dpt;
+		*dpt = ".";
+		return 1;
+	}
+	**npt = 0;
+	if(*npt == *dpt) 
+		*dpt = "/";
+	(*npt)++;
+	return 2;
+}
+
+
+/* ts A70923 : API */
+int burn_drive_equals_adr(struct burn_drive *d1, char *adr2_in, int role2)
+{
+	struct stat stbuf1, stbuf2;
+	char adr1[BURN_DRIVE_ADR_LEN], *adr2 = adr2_in;
+	char conv_adr1[BURN_DRIVE_ADR_LEN], conv_adr2[BURN_DRIVE_ADR_LEN];
+	char *npt1, *dpt1, *npt2, *dpt2;
+	int role1, stat_ret1, stat_ret2, conv_ret2;
+
+	role1 = burn_drive_get_drive_role(d1);
+	burn_drive_d_get_adr(d1, adr1);
+	stat_ret1 = stat(adr1, &stbuf1);
+
+	if (strncmp(adr2, "stdio:", 6) == 0) {
+		adr2+= 6;
+		role2 = (!!adr2[0]) * 2;
+	}
+	if (strlen(adr2) >= BURN_DRIVE_ADR_LEN)
+		return -1;
+	stat_ret2 = stat(adr2, &stbuf2);
+	conv_ret2 = burn_drive_convert_fs_adr(adr2, conv_adr2);
+
+	/* roles 2 and 3 have the same name space and object interpretation */
+	if (role1 == 3)
+		role1 = 2;
+	if (role2 == 3)
+		role2 = 2;
+
+	if (strcmp(adr1, adr2) == 0 && role1 == role2)
+		return(1);			/* equal role and address */
+	if (role1 == 1 && role2 == 1) {
+					/* MMC drive meets wannabe MMC drive */
+		if (conv_ret2 <= 0)
+			return 0;		/* no MMC drive at adr2 */
+		if (strcmp(adr1, conv_adr2) == 0)
+			return 1;		/* equal real MMC drives */
+		return 0;
+
+	} else if (role1 == 0 || role2 == 0)
+		return 0;			/* one null-drive, one not */
+
+	else if (role1 != 1 && role2 != 1) {
+					/* pseudo-drive meets file object */
+
+		if (stat_ret1 == -1 || stat_ret2 == -1) {
+			if (stat_ret1 != -1 || stat_ret2 != -1)
+				 return 0;  /* one adress existing, one not */
+
+			/* Two non-existing file objects */
+
+			strcpy(conv_adr1, adr1);
+			burn__split_path(conv_adr1, &dpt1, &npt1);
+			strcpy(conv_adr2, adr2);
+			burn__split_path(conv_adr2, &dpt2, &npt2);
+			if (strcmp(npt1, npt2))
+				return 0;		/* basenames differ */
+			stat_ret1= stat(adr1, &stbuf1);
+			stat_ret2= stat(adr2, &stbuf2);
+			if (stat_ret1 != stat_ret2)
+				 return 0;  /* one dir existing, one not */
+
+			/* Both directories exist. The basenames are equal.
+			   So the adresses are equal if the directories are
+			   equal.*/
+		}
+		if (stbuf1.st_ino == stbuf2.st_ino &&
+	 		stbuf1.st_dev == stbuf2.st_dev)
+			return 1;		/* same filesystem object */
+
+		if (S_ISBLK(stbuf1.st_mode) && S_ISBLK(stbuf2.st_mode) &&
+			 stbuf1.st_rdev == stbuf2.st_rdev)
+			return 1;	/* same major,minor device number */
+
+		/* Are both filesystem objects related to the same MMC drive */
+		if (conv_ret2 <= 0)
+			return 0;		/* no MMC drive at adr2 */
+		if (burn_drive_convert_fs_adr(adr1, conv_adr1) <= 0)
+			return 0;		/* no MMC drive at adr1 */
+		if (strcmp(conv_adr1, conv_adr2) == 0)
+			return 1;		/* same MMC drive */
+
+		return 0;	/* all filesystem disguises are checked */
+
+	} else if (role1 == 1 && role2 != 1) {
+	                          /* MMC drive meets file object */
+
+		if (conv_ret2 <= 0)
+			return 0;		/* no MMC drive at adr2 */
+		if (strcmp(adr1, conv_adr2) == 0)
+			return 1;		/* same MMC drive */
+		return 0;
+
+	} else if (role1 != 1 && role2 == 1) {
+	                          /* stdio-drive meets wannabe MMC drive */
+
+		if (conv_ret2 <= 0)
+			return 0;		/* no MMC drive at adr2 */
+		if (burn_drive_convert_fs_adr(adr1, conv_adr1) <= 0)
+			return 0;		/* no MMC drive at adr1 */
+		if (strcmp(conv_adr1, conv_adr2) == 0)
+			return 1;		/* same MMC drive */
+		return 0;
+
+	}
+	return 0;		/* now i believe they are really not equal */
+}
+
+
+int burn_drive_find_by_thread_pid(struct burn_drive **d, pid_t pid)
+{
+	int i;
+
+	for (i = 0; i < drivetop + 1; i++) {
+
+/*
+		if (drive_array[i].thread_pid_valid)
+			fprintf(stderr, "libburn_EXPERIMENTAL : drive %d , thread_pid %d\n", i, drive_array[i].thread_pid);
+*/
+
+		if (drive_array[i].thread_pid_valid &&
+		    drive_array[i].thread_pid == pid) {
+			*d = &(drive_array[i]);
+			return 1;
+		}
+	}
+	return 0;
+}
+
+
