@@ -8,12 +8,18 @@
 #include <stdio.h>
 #include <signal.h>
 #include <string.h>
+#include <stdlib.h>
 
+/* ts A70928 : init.h is for others, not for init .c
 #include "init.h"
+*/
+
+
 #include "sg.h"
 #include "error.h"
 #include "libburn.h"
 #include "drive.h"
+#include "transport.h"
 
 /* ts A60825 : The storage location for back_hacks.h variables. */
 #define BURN_BACK_HACKS_INIT 1
@@ -53,7 +59,6 @@ int burn_sg_open_o_nonblock = 1;
 	    to unconditional abort of the process  */
 int burn_sg_open_abort_busy = 0;
 
-
 /* ts A61002 */
 
 #include "cleanup.h"
@@ -61,6 +66,10 @@ int burn_sg_open_abort_busy = 0;
 /* Parameters for builtin abort handler */
 static char abort_message_prefix[81] = {"libburn : "};
 static pid_t abort_control_pid= 0;
+volatile int burn_global_abort_level= 0;
+int burn_global_abort_signum= 0;
+void *burn_global_signal_handle = NULL;
+burn_abort_handler_t burn_global_signal_handler = NULL;
 
 
 /* ts A70223 : wether implemented untested profiles are supported */
@@ -107,11 +116,12 @@ void burn_finish(void)
 
 	/* ts A61007 */
 	/* burn_wait_all(); */
-	if (!burn_drives_are_clear()) {
+	if (!burn_drives_are_clear(0)) {
 		libdax_msgs_submit(libdax_messenger, -1, 0x00020107,
-			LIBDAX_MSGS_SEV_SORRY, LIBDAX_MSGS_PRIO_HIGH,
-			"Drive is busy on attempt to shut down library", 0, 0);
-		return;
+			LIBDAX_MSGS_SEV_WARNING, LIBDAX_MSGS_PRIO_HIGH,
+			"A drive is still busy on shutdown of library", 0, 0);
+		usleep(1000001);
+		burn_abort(4440, burn_abort_pacifier, abort_message_prefix);
 	}
 
 	/* ts A60904 : ticket 62, contribution by elmom : name addon "_all" */
@@ -200,6 +210,8 @@ int burn_msgs_obtain(char *minimum_severity,
 	ret = libdax_msgs__text_to_sev(minimum_severity, &minimum_sevno, 0);
 	if (ret <= 0)
 		return 0;
+	if (libdax_messenger == NULL)
+		return 0;
 	ret = libdax_msgs_obtain(libdax_messenger, &item, minimum_sevno,
 				LIBDAX_MSGS_PRIO_ZERO, 0);
 	if (ret <= 0)
@@ -226,40 +238,116 @@ ex:
 	return ret;
 }
 
+
+/* ts A70922 : API */
+int burn_msgs_submit(int error_code, char msg_text[], int os_errno,
+			char severity[], struct burn_drive *d)
+{
+	int ret, sevno, global_index = -1;
+
+	ret = libdax_msgs__text_to_sev(severity, &sevno, 0);
+	if (ret <= 0)
+		sevno = LIBDAX_MSGS_SEV_FATAL;
+	if (error_code <= 0) {
+		switch(sevno) {
+		       case LIBDAX_MSGS_SEV_ABORT:   error_code = 0x00040000;
+		break; case LIBDAX_MSGS_SEV_FATAL:   error_code = 0x00040001;
+		break; case LIBDAX_MSGS_SEV_SORRY:   error_code = 0x00040002;
+		break; case LIBDAX_MSGS_SEV_WARNING: error_code = 0x00040003;
+		break; case LIBDAX_MSGS_SEV_HINT:    error_code = 0x00040004;
+		break; case LIBDAX_MSGS_SEV_NOTE:    error_code = 0x00040005;
+		break; case LIBDAX_MSGS_SEV_UPDATE:  error_code = 0x00040006;
+		break; case LIBDAX_MSGS_SEV_DEBUG:   error_code = 0x00040007;
+		break; default:                      error_code = 0x00040001;
+		}
+	}
+	if (d != NULL)
+		global_index = d->global_index;
+	ret = libdax_msgs_submit(libdax_messenger, global_index, error_code,
+			sevno, LIBDAX_MSGS_PRIO_HIGH, msg_text, os_errno, 0);
+	return ret;
+}
+
+
+/* ts A71016 API */
+int burn_text_to_sev(char *severity_name, int *sevno, int flag)
+{
+	int ret;
+
+	ret = libdax_msgs__text_to_sev(severity_name, sevno, 0);
+	if (ret <= 0)
+		*sevno = LIBDAX_MSGS_SEV_FATAL;
+	return ret;
+}
+
+
 int burn_builtin_abort_handler(void *handle, int signum, int flag)
 {
+
+#define Libburn_new_thread_signal_handleR 1
+/*
+#define Libburn_signal_handler_verbouS 1
+*/
+	int ret;
+	struct burn_drive *d;
+
+#ifdef Libburn_signal_handler_verbouS
+	fprintf(stderr,
+		"libburn_ABORT: pid = %d , abort_control_pid = %d , sig= %d\n",
+		getpid(), abort_control_pid, signum);
+#endif
+
+	/* ts A70928:
+	Must be quick. Allowed to coincide with other thread and to share
+	the increment with that one. It must not decrease, though, and
+	yield at least 1 if any thread calls this function.
+	*/
+	burn_global_abort_level++;
+	burn_global_abort_signum= signum;
+
 	if(getpid() != abort_control_pid) {
 
-#ifdef Not_yeT
-		pthread_t thread_id;
+#ifdef Libburn_new_thread_signal_handleR
 
-		/* >>> need better handling of self-induced SIGs 
-		       like SIGSEGV or SIGFPE.
-		       Like bonking the control thread if it did not show up
-		       after a short while.
-		*/
+		ret = burn_drive_find_by_thread_pid(&d, getpid());
+		if (ret > 0 && d->busy == BURN_DRIVE_WRITING) {
+					/* This is an active writer thread */
 
-		/* >>> if this is a non-fatal signal : return -2 */
+#ifdef Libburn_signal_handler_verbouS
+			fprintf(stderr, "libburn_ABORT: pid %d found drive busy with writing, (level= %d)\n", (int) getpid(), burn_global_abort_level);
+#endif
 
-		thread_id = pthread_self();
-		/* >>> find thread_id  in worker list of async.c */
-		/* >>> if owning a drive : mark idle and canceled
-		       (can't do anything more) */
+			d->sync_cache(d);
 
-		usleep(1000000); /* calm down */
+			/* >>> perform a more qualified end of burn process */;
 
- 		/* forward signal to control thread */
-		if (abort_control_pid>1)
-			kill(abort_control_pid, signum);
+			d->busy = BURN_DRIVE_IDLE;
 
-		/* >>> ??? end thread */;
+			if (burn_global_abort_level > 0) {
+				/* control process did not show up yet */
+#ifdef Libburn_signal_handler_verbouS
+					fprintf(stderr, "libburn_ABORT: pid %d sending signum %d to pid %d\n", (int) getpid(), (int) signum, (int) abort_control_pid);
+#endif
+					kill(abort_control_pid, signum);
+			}
+
+#ifdef Libburn_signal_handler_verbouS
+					fprintf(stderr, "libburn_ABORT: pid %d signum %d returning -2\n", (int) getpid(), (int) signum);
+#endif
+
+			return -2;
+		} else {
+			usleep(1000000); /* calm down */
+			return -2;
+		}
 
 #else
 		usleep(1000000); /* calm down */
 		return -2;
-#endif /* ! Not_yeT */
+#endif /* ! Libburn_new_thread_signal_handleR */
 
 	}
+	burn_global_abort_level = -1;
 	Cleanup_set_handlers(NULL, NULL, 2);
 	fprintf(stderr,"%sABORT : Trying to shut down drive and library\n",
 		abort_message_prefix);
@@ -267,10 +355,13 @@ int burn_builtin_abort_handler(void *handle, int signum, int flag)
 		"%sABORT : Wait the normal burning time before any kill -9\n",
 		abort_message_prefix);
 	close(0); /* somehow stdin as input blocks abort until EOF */
+
 	burn_abort(4440, burn_abort_pacifier, abort_message_prefix);
+
 	fprintf(stderr,
 	"\n%sABORT : Program done. Even if you do not see a shell prompt.\n\n",
 		abort_message_prefix);
+	burn_global_abort_level = -2;
 	return(1);
 }
 
@@ -289,8 +380,10 @@ void burn_set_signal_handling(void *handle, burn_abort_handler_t handler,
 		strncpy(abort_message_prefix, (char *) handle,
 			sizeof(abort_message_prefix)-1);
 	abort_message_prefix[sizeof(abort_message_prefix)-1] = 0;
-	abort_control_pid= getpid();
+	abort_control_pid = getpid();
 	Cleanup_set_handlers(handle, (Cleanup_app_handler_T) handler, mode|4);
+	burn_global_signal_handle = handle;
+	burn_global_signal_handler = handler;
 }
 
 
@@ -300,3 +393,16 @@ void burn_allow_untested_profiles(int yes)
 	burn_support_untested_profiles = !!yes;
 }
 
+
+/* ts A70915 : API */
+int burn_set_messenger(void *messenger)
+{
+	struct libdax_msgs *pt;
+
+	if (libdax_msgs_refer(&pt, messenger, 0) <= 0)
+		return 0;
+	libdax_msgs_destroy(&libdax_messenger, 0);
+	libdax_messenger = (struct libdax_msgs *) pt;
+	return 1;
+}
+
