@@ -116,6 +116,7 @@ extern struct libdax_msgs *libdax_messenger;
 static unsigned char MMC_GET_MSINFO[] =
 	{ 0x43, 0, 1, 0, 0, 0, 0, 16, 0, 0 };
 static unsigned char MMC_GET_TOC[] = { 0x43, 2, 2, 0, 0, 0, 0, 16, 0, 0 };
+static unsigned char MMC_GET_TOC_FMT0[] = { 0x43, 0, 0, 0, 0, 0, 0, 16, 0, 0 };
 static unsigned char MMC_GET_ATIP[] = { 0x43, 2, 4, 0, 0, 0, 0, 16, 0, 0 };
 static unsigned char MMC_GET_DISC_INFO[] =
 	{ 0x51, 0, 0, 0, 0, 0, 0, 16, 0, 0 };
@@ -165,15 +166,32 @@ static unsigned char MMC_READ_FORMAT_CAPACITIES[] =
 static unsigned char MMC_RESERVE_TRACK[] =
 	{ 0x53, 0, 0, 0, 0, 0, 0, 0, 0, 0};
 
+/* ts A70812 : Read data sectors (for types with 2048 bytes/sector only) */
+static unsigned char MMC_READ_10[] =
+	{ 0x28, 0, 0, 0, 0, 0, 0, 0, 0, 0};
+
 
 static int mmc_function_spy_do_tell = 0;
 
-int mmc_function_spy(char * text)
+int mmc_function_spy(struct burn_drive *d, char * text)
 {
-
 	if (mmc_function_spy_do_tell)
 		fprintf(stderr,"libburn: experimental: mmc_function_spy: %s\n",
 			text);
+	if (d == NULL)
+		return 1;
+	if (d->drive_role != 1) {
+		char msg[4096];
+	
+		sprintf(msg, "Emulated drive caught in SCSI adapter \"%s\"",
+				text);
+		libdax_msgs_submit(libdax_messenger, d->global_index,
+				0x0002014c,
+				LIBDAX_MSGS_SEV_FATAL, LIBDAX_MSGS_PRIO_HIGH,
+				msg, 0, 0);
+		d->cancel = 1;
+		return 0;
+	}
 	return 1;
 }
 
@@ -208,7 +226,8 @@ void mmc_send_cue_sheet(struct burn_drive *d, struct cue_sheet *s)
 	struct command c;
 
 
-	mmc_function_spy("mmc_send_cue_sheet");
+	if (mmc_function_spy(d, "mmc_send_cue_sheet") <= 0)
+		return;
 
 	scsi_init_command(&c, MMC_SEND_CUE_SHEET, sizeof(MMC_SEND_CUE_SHEET));
 /*
@@ -238,7 +257,9 @@ int mmc_reserve_track(struct burn_drive *d, off_t size)
 	int lba;
 	char msg[80];
 
-	mmc_function_spy("mmc_reserve_track");
+	if (mmc_function_spy(d, "mmc_reserve_track") <= 0)
+		return 0;
+
 	scsi_init_command(&c, MMC_RESERVE_TRACK, sizeof(MMC_RESERVE_TRACK));
 /*
 	c.oplen = sizeof(MMC_RESERVE_TRACK);
@@ -270,7 +291,8 @@ int mmc_read_track_info(struct burn_drive *d, int trackno, struct buffer *buf,
 {
 	struct command c;
 
-	mmc_function_spy("mmc_read_track_info");
+	if (mmc_function_spy(d, "mmc_read_track_info") <= 0)
+		return 0;
 
 	scsi_init_command(&c, MMC_TRACK_INFO, sizeof(MMC_TRACK_INFO));
 /*
@@ -316,7 +338,8 @@ int mmc_get_nwa(struct burn_drive *d, int trackno, int *lba, int *nwa)
 	int ret, num, alloc_len = 20;
 	unsigned char *data;
 
-	mmc_function_spy("mmc_get_nwa");
+	if (mmc_function_spy(d, "mmc_get_nwa") <= 0)
+		return -1;
 
 	ret = mmc_read_track_info(d, trackno, &buf, alloc_len);
 	if (ret <= 0)
@@ -354,9 +377,10 @@ int mmc_get_nwa(struct burn_drive *d, int trackno, int *lba, int *nwa)
 /* void mmc_close_disc(struct burn_drive *d, struct burn_write_opts *o) */
 void mmc_close_disc(struct burn_write_opts *o)
 {
-	struct burn_drive *d;
+	struct burn_drive *d = o->drive;
 
-	mmc_function_spy("mmc_close_disc");
+	if (mmc_function_spy(d, "mmc_close_disc") <= 0)
+		return;
 
 	libdax_msgs_submit(libdax_messenger, -1, 0x00000002,
 			   LIBDAX_MSGS_SEV_DEBUG, LIBDAX_MSGS_PRIO_ZERO,
@@ -364,7 +388,6 @@ void mmc_close_disc(struct burn_write_opts *o)
 
 	/* ts A61009 : made impossible by removing redundant parameter d */
 	/* a ssert(o->drive == d); */
-	d = o->drive;
 
 	o->multi = 0;
 	spc_select_write_params(d, o);
@@ -375,9 +398,10 @@ void mmc_close_disc(struct burn_write_opts *o)
 /* void mmc_close_session(struct burn_drive *d, struct burn_write_opts *o) */
 void mmc_close_session(struct burn_write_opts *o)
 {
-	struct burn_drive *d;
+	struct burn_drive *d = o->drive;
 
-	mmc_function_spy("mmc_close_session");
+	if (mmc_function_spy(d, "mmc_close_session") <= 0)
+		return;
 
 	libdax_msgs_submit(libdax_messenger, -1, 0x00000002,
 			   LIBDAX_MSGS_SEV_DEBUG, LIBDAX_MSGS_PRIO_ZERO,
@@ -385,7 +409,6 @@ void mmc_close_session(struct burn_write_opts *o)
 
 	/* ts A61009 : made impossible by removing redundant parameter d */
 	/* a ssert(o->drive == d); */
-	d = o->drive;
 
 	o->multi = 3;
 	spc_select_write_params(d, o);
@@ -401,7 +424,8 @@ void mmc_close(struct burn_drive *d, int session, int track)
 {
 	struct command c;
 
-	mmc_function_spy("mmc_close");
+	if (mmc_function_spy(d, "mmc_close") <= 0)
+		return;
 
 	scsi_init_command(&c, MMC_CLOSE, sizeof(MMC_CLOSE));
 /*
@@ -410,6 +434,8 @@ void mmc_close(struct burn_drive *d, int session, int track)
 */
 	c.retry = 1;
 
+	c.opcode[1] |= 1; /* ts A70918 : Immed */
+
 	/* (ts A61030 : shifted !!session rather than or-ing plain session ) */
 	c.opcode[2] = ((session & 3) << 1) | !!track;
 	c.opcode[4] = track >> 8;
@@ -417,6 +443,14 @@ void mmc_close(struct burn_drive *d, int session, int track)
 	c.page = NULL;
 	c.dir = NO_TRANSFER;
 	d->issue_command(d, &c);
+
+	/* ts A70918 : Immed : wait for drive to complete command */
+	if (c.error) {
+		d->cancel = 1;
+		return;
+	}
+	if (spc_wait_unit_attention(d, 3600, "CLOSE TRACK SESSION", 0) <= 0)
+		d->cancel = 1;
 }
 
 void mmc_get_event(struct burn_drive *d)
@@ -425,7 +459,8 @@ void mmc_get_event(struct burn_drive *d)
 	struct command c;
 	int alloc_len= 8;
 
-	mmc_function_spy("mmc_get_event");
+	if (mmc_function_spy(d, "mmc_get_event") <= 0)
+		return;
 
 	scsi_init_command(&c, MMC_GET_EVENT, sizeof(MMC_GET_EVENT));
 /*
@@ -495,7 +530,7 @@ static int mmc_wait_for_buffer_free(struct burn_drive *d, struct buffer *buf)
 	/* There is need to inquire the buffer fill */
 	d->pessimistic_writes++;
 	min_fac = ((double) d->wfb_min_percent) / 100.0;
-	gettimeofday(&t0,&dummy_tz);
+	gettimeofday(&t0, &dummy_tz);
 #ifdef Libburn_mmc_wfb_debuG
 	sleeplist[0]= 0;
 	sprintf(sleeplist,"(%d%s %d)",
@@ -590,7 +625,9 @@ void mmc_write_12(struct burn_drive *d, int start, struct buffer *buf)
 	struct command c;
 	int len;
 
-	mmc_function_spy("mmc_write_12");
+	if (mmc_function_spy(d, "mmc_write_12") <= 0)
+		return;
+
 	len = buf->sectors;
 
 	/* ts A61009 */
@@ -630,11 +667,10 @@ int mmc_write(struct burn_drive *d, int start, struct buffer *buf)
 				O_WRONLY|O_CREAT|O_TRUNC,S_IRUSR|S_IWUSR);
 #endif /* Libburn_log_in_and_out_streaM */
 
-	mmc_function_spy("mmc_write");
-	pthread_mutex_lock(&d->access_lock);
-	cancelled = d->cancel;
-	pthread_mutex_unlock(&d->access_lock);
+	if (mmc_function_spy(d, "mmc_write") <= 0)
+		return BE_CANCELLED;
 
+	cancelled = d->cancel;
 	if (cancelled)
 		return BE_CANCELLED;
 
@@ -709,9 +745,7 @@ int mmc_write(struct burn_drive *d, int start, struct buffer *buf)
 				LIBDAX_MSGS_SEV_FATAL, LIBDAX_MSGS_PRIO_HIGH,
 				msg, 0, 0);
 		}
-		pthread_mutex_lock(&d->access_lock);
 		d->cancel = 1;
-		pthread_mutex_unlock(&d->access_lock);
 		return BE_CANCELLED;
 	} 
 
@@ -764,6 +798,153 @@ int mmc_fake_toc_entry(struct burn_toc_entry *entry, int session_number,
 }
 
 
+/* ts A71128 : for DVD-ROM drives which offer no reliable track information */
+static int mmc_read_toc_fmt0_al(struct burn_drive *d, int *alloc_len)
+{
+	struct burn_track *track;
+	struct burn_session *session;
+	struct burn_toc_entry *entry;
+	struct buffer buf;
+	struct command c;
+	int dlen, i, old_alloc_len, session_number, prev_session = -1;
+	int lba, size;
+	unsigned char *tdata, size_data[4], start_data[4];
+
+	if (*alloc_len < 4)
+		return 0;
+
+	scsi_init_command(&c, MMC_GET_TOC_FMT0, sizeof(MMC_GET_TOC_FMT0));
+	c.dxfer_len = *alloc_len;
+	c.opcode[7] = (c.dxfer_len >> 8) & 0xff;
+	c.opcode[8] = c.dxfer_len & 0xff;
+	c.retry = 1;
+	c.page = &buf;
+	c.page->bytes = 0;
+	c.page->sectors = 0;
+	c.dir = FROM_DRIVE;
+	d->issue_command(d, &c);
+
+	if (c.error) {
+err_ex:;
+		libdax_msgs_submit(libdax_messenger, d->global_index,
+			 0x0002010d,
+			 LIBDAX_MSGS_SEV_DEBUG, LIBDAX_MSGS_PRIO_HIGH,
+			 "Could not inquire TOC", 0,0);
+		d->status = BURN_DISC_UNSUITABLE;
+		d->toc_entries = 0;
+		/* Prefering memory leaks over fandangos */
+		d->toc_entry = calloc(1, sizeof(struct burn_toc_entry));
+		return 0;
+	}
+	dlen = c.page->data[0] * 256 + c.page->data[1];
+	old_alloc_len = *alloc_len;
+	*alloc_len = dlen + 2;
+	if (old_alloc_len < 12)
+		return 1;
+	if (dlen + 2 > old_alloc_len)
+		dlen = old_alloc_len - 2;
+	d->complete_sessions = 1 + c.page->data[3] - c.page->data[2];
+	d->last_track_no = d->complete_sessions;
+	if (dlen - 2 < (d->last_track_no + 1) * 8) {
+		libdax_msgs_submit(libdax_messenger, d->global_index,
+			 0x00020159,
+			 LIBDAX_MSGS_SEV_DEBUG, LIBDAX_MSGS_PRIO_HIGH,
+			 "TOC Format 0 returns inconsistent data", 0,0);
+		goto err_ex;
+	}
+
+	d->toc_entries = d->last_track_no + d->complete_sessions;
+	if (d->toc_entries < 1)
+		return 0;
+	d->toc_entry = calloc(d->toc_entries, sizeof(struct burn_toc_entry));
+	if(d->toc_entry == NULL)
+		return 0;
+
+	d->disc = burn_disc_create();
+	if (d->disc == NULL)
+		return 0;
+	for (i = 0; i < d->complete_sessions; i++) {
+		session = burn_session_create();
+		if (session == NULL)
+			return 0;
+		burn_disc_add_session(d->disc, session, BURN_POS_END);
+		burn_session_free(session);
+	}
+
+
+	for (i = 0; i < d->last_track_no; i++) {
+		tdata = c.page->data + 4 + i * 8;
+		session_number = i + 1;
+		if (session_number != prev_session && prev_session > 0) {
+			/* leadout entry previous session */
+			entry = &(d->toc_entry[(i - 1) + prev_session]);
+			lba = mmc_four_char_to_int(start_data) +
+			      mmc_four_char_to_int(size_data);
+			mmc_int_to_four_char(start_data, lba);
+			mmc_int_to_four_char(size_data, 0);
+			mmc_fake_toc_entry(entry, prev_session, 0xA2,
+					 size_data, start_data);
+			entry->min= entry->sec= entry->frame= 0;
+			d->disc->session[prev_session - 1]->leadout_entry =
+									entry;
+		}
+
+		/* ??? >>> d->media_capacity_remaining , d->media_lba_limit
+				as of mmc_fake_toc()
+		*/
+
+		entry = &(d->toc_entry[i + session_number - 1]);
+ 		track = burn_track_create();
+		if (track == NULL)
+			return -1;
+		burn_session_add_track(
+			d->disc->session[session_number - 1],
+			track, BURN_POS_END);
+		track->entry = entry;
+		burn_track_free(track);
+
+		memcpy(start_data, tdata + 4, 4);
+			/* size_data are estimated from next track start */
+		memcpy(size_data, tdata + 8 + 4, 4);
+		size = mmc_four_char_to_int(size_data) -
+	      		mmc_four_char_to_int(start_data);
+		mmc_int_to_four_char(size_data, size);
+		mmc_fake_toc_entry(entry, session_number, i + 1,
+					 size_data, start_data);
+		if (prev_session != session_number)
+			d->disc->session[session_number - 1]->firsttrack = i+1;
+		d->disc->session[session_number - 1]->lasttrack = i+1;
+		prev_session = session_number;
+	}
+	if (prev_session > 0 && prev_session <= d->disc->sessions) {
+		/* leadout entry of last session of closed disc */
+		tdata = c.page->data + 4 + d->last_track_no * 8;
+		entry = &(d->toc_entry[(d->last_track_no - 1) + prev_session]);
+		memcpy(start_data, tdata + 4, 4);
+		mmc_int_to_four_char(size_data, 0);
+		mmc_fake_toc_entry(entry, prev_session, 0xA2,
+				 size_data, start_data);
+		entry->min= entry->sec= entry->frame= 0;
+		d->disc->session[prev_session - 1]->leadout_entry = entry;
+	}
+	return 1;
+}
+
+
+/* ts A71128 : for DVD-ROM drives which offer no reliable track information */
+static int mmc_read_toc_fmt0(struct burn_drive *d)
+{
+	int alloc_len = 4, ret;
+
+	if (mmc_function_spy(d, "mmc_read_toc_fmt0") <= 0)
+		return -1;
+	ret = mmc_read_toc_fmt0_al(d, &alloc_len);
+	if (alloc_len >= 12)
+		ret = mmc_read_toc_fmt0_al(d, &alloc_len);
+	return ret;
+}
+
+
 /* ts A70131 : compose a disc TOC structure from d->complete_sessions
                and 52h READ TRACK INFORMATION */
 int mmc_fake_toc(struct burn_drive *d)
@@ -776,6 +957,9 @@ int mmc_fake_toc(struct burn_drive *d)
 	unsigned char *tdata, size_data[4], start_data[4];
 	char msg[160];
 
+	if (mmc_function_spy(d, "mmc_fake_toc") <= 0)
+		return -1;
+
 	if (d->last_track_no <= 0 || d->complete_sessions <= 0 ||
 	    d->status == BURN_DISC_BLANK)
 		return 2;
@@ -789,6 +973,12 @@ int mmc_fake_toc(struct burn_drive *d)
 				 msg, 0,0);
 		return 0;
 	}
+	/* ts A71128 : My DVD-ROM drive issues no reliable track info.
+			One has to try 43h READ TOC/PMA/ATIP Form 0. */
+	if (d->current_profile == 0x10 && d->last_track_no <= 1) {
+		ret = mmc_read_toc_fmt0(d);
+		return ret;
+	}
 	d->disc = burn_disc_create();
 	if (d->disc == NULL)
 		return -1;
@@ -799,12 +989,15 @@ int mmc_fake_toc(struct burn_drive *d)
 	memset(d->toc_entry, 0,d->toc_entries * sizeof(struct burn_toc_entry));
 	for (i = 0; i < d->complete_sessions; i++) {
 		session = burn_session_create();
+		if (session == NULL)
+			return -1;
 		burn_disc_add_session(d->disc, session, BURN_POS_END);
 		burn_session_free(session);
 	}
 	memset(size_data, 0, 4);
 	memset(start_data, 0, 4);
 
+
 	/* Entry Layout :
 		session 1   track 1     entry 0
 		...
@@ -901,21 +1094,21 @@ static int mmc_read_toc_al(struct burn_drive *d, int *alloc_len)
 	int i, bpl= 12, old_alloc_len;
 	unsigned char *tdata;
 
-	mmc_function_spy("mmc_read_toc");
-
 	if (*alloc_len < 4)
 		return 0;
 
 	if (!(d->current_profile == -1 || d->current_is_cd_profile)) {
 		/* ts A70131 : MMC_GET_TOC uses Response Format 2 
 		   For DVD this fails with 5,24,00 */
-		/* One could try Response Format 0: mmc5r03.pdf 6.26.3.2
-		   which does not yield the same result wit the same disc
+		/* mmc_read_toc_fmt0() uses
+                   Response Format 0: mmc5r03.pdf 6.26.3.2
+		   which does not yield the same result with the same disc
 		   on different drives.
 		*/
 		/* ts A70201 :
 		   This uses the session count from 51h READ DISC INFORMATION
-		   and the track records from 52h READ TRACK INFORMATION
+		   and the track records from 52h READ TRACK INFORMATION.
+		   mmc_read_toc_fmt0() is used as fallback for dull DVD-ROM.
 		*/
 		mmc_fake_toc(d);
 
@@ -955,9 +1148,7 @@ static int mmc_read_toc_al(struct burn_drive *d, int *alloc_len)
 		d->status = BURN_DISC_UNSUITABLE;
 		d->toc_entries = 0;
 		/* Prefering memory leaks over fandangos */
-		d->toc_entry = malloc(sizeof(struct burn_toc_entry));
-		memset(&(d->toc_entry[0]), 0, sizeof(struct burn_toc_entry));
-
+		d->toc_entry = calloc(1, sizeof(struct burn_toc_entry));
 		return 0;
 	}
 
@@ -977,17 +1168,21 @@ static int mmc_read_toc_al(struct burn_drive *d, int *alloc_len)
 	ts A61007 : if re-enabled then not via Assert.
 	a ssert(((dlen - 2) % 11) == 0);
 */
-	d->toc_entry = malloc(d->toc_entries * sizeof(struct burn_toc_entry));
-	for (i = 0; i < d->toc_entries; i++)
-		memset(&(d->toc_entry[i]), 0, sizeof(struct burn_toc_entry));
+	d->toc_entry = calloc(d->toc_entries, sizeof(struct burn_toc_entry));
+	if(d->toc_entry == NULL) /* ts A70825 */
+		return 0;
 	tdata = c.page->data + 4;
 
 	burn_print(12, "TOC:\n");
 
 	d->disc = burn_disc_create();
+	if (d->disc == NULL) /* ts A70825 */
+		return 0;
 
 	for (i = 0; i < c.page->data[3]; i++) {
 		session = burn_session_create();
+		if (session == NULL) /* ts A70825 */
+			return 0;
 		burn_disc_add_session(d->disc, session, BURN_POS_END);
 		burn_session_free(session);
 	}
@@ -1067,6 +1262,9 @@ void mmc_read_toc(struct burn_drive *d)
 {
 	int alloc_len = 4, ret;
 
+	if (mmc_function_spy(d, "mmc_read_toc") <= 0)
+		return;
+
 	ret = mmc_read_toc_al(d, &alloc_len);
 /*
 	fprintf(stderr,
@@ -1094,7 +1292,8 @@ int mmc_read_multi_session_c1(struct burn_drive *d, int *trackno, int *start)
 	struct burn_track **tracks;
 	struct burn_toc_entry toc_entry;
 
-	mmc_function_spy("mmc_read_multi_session_c1");
+	if (mmc_function_spy(d, "mmc_read_multi_session_c1") <= 0)
+		return 0;
 
 	/* First try to evaluate the eventually loaded TOC before issueing
 	   a MMC command. This search obtains the first track of the last
@@ -1181,9 +1380,9 @@ static int mmc_read_disc_info_al(struct burn_drive *d, int *alloc_len)
 
 	mmc_get_configuration(d);
 
+/* ts A70910 : found this as condition for mmc_function_spy() which went up
 	if (*alloc_len < 2)
-
-	mmc_function_spy("mmc_read_disc_info");
+*/
 
 	scsi_init_command(&c, MMC_GET_DISC_INFO, sizeof(MMC_GET_DISC_INFO));
 /*
@@ -1211,7 +1410,7 @@ static int mmc_read_disc_info_al(struct burn_drive *d, int *alloc_len)
 	*alloc_len = len + 2;
 	if (old_alloc_len < 34)
 		return 1;
-	if (*alloc_len < 24) /* data[23] is the last byte used her */
+	if (*alloc_len < 24) /* data[23] is the last byte used here */
 		return 0;
 	if (len + 2 > old_alloc_len)
 		len = old_alloc_len - 2;
@@ -1300,6 +1499,9 @@ static int mmc_read_disc_info_al(struct burn_drive *d, int *alloc_len)
 		   appendable. I.e number of complete tracks + 1. */
 		d->last_track_no = (data[11] << 8) | data[6];
 	}
+	if (d->current_profile != 0x0a && d->current_profile != 0x13 &&
+	    d->current_profile != 0x14 && d->status != BURN_DISC_FULL)
+		d->erasable = 0; /* stay in sync with burn_disc_erase() */
 
 	if (do_read_toc)
 		mmc_read_toc(d);
@@ -1311,6 +1513,9 @@ void mmc_read_disc_info(struct burn_drive *d)
 {
 	int alloc_len = 34, ret;
 
+	if (mmc_function_spy(d, "mmc_read_disc_info") <= 0)
+		return;
+
 	ret = mmc_read_disc_info_al(d, &alloc_len);
 /*
 	fprintf(stderr,"LIBBURN_DEBUG: 51h alloc_len = %d , ret = %d\n",
@@ -1339,7 +1544,8 @@ void mmc_read_atip(struct burn_drive *d)
 	                   4234, 5646, 7056, 8468, -12, -13, -14, -15};
 	               /*    24,   32,   40,   48,   -,   -,   -,   - */
 
-	mmc_function_spy("mmc_read_atip");
+	if (mmc_function_spy(d, "mmc_read_atip") <= 0)
+		return;
 
 	scsi_init_command(&c, MMC_GET_ATIP, sizeof(MMC_GET_ATIP));
 /*
@@ -1495,7 +1701,8 @@ void mmc_read_sectors(struct burn_drive *d,
 	int errorblock, req;
 	struct command c;
 
-	mmc_function_spy("mmc_read_sectors");
+	if (mmc_function_spy(d, "mmc_read_sectors") <= 0)
+		return;
 
 	/* ts A61009 : to be ensured by callers */
 	/* a ssert(len >= 0); */
@@ -1564,7 +1771,8 @@ void mmc_erase(struct burn_drive *d, int fast)
 {
 	struct command c;
 
-	mmc_function_spy("mmc_erase");
+	if (mmc_function_spy(d, "mmc_erase") <= 0)
+		return;
 
 	scsi_init_command(&c, MMC_BLANK, sizeof(MMC_BLANK));
 /*
@@ -1584,7 +1792,9 @@ void mmc_read_lead_in(struct burn_drive *d, struct buffer *buf)
 	int len;
 	struct command c;
 
-	mmc_function_spy("mmc_read_lead_in");
+	if (mmc_function_spy(d, "mmc_read_lead_in") <= 0)
+		return;
+
 	len = buf->sectors;
 	scsi_init_command(&c, MMC_READ_CD, sizeof(MMC_READ_CD));
 /*
@@ -1610,7 +1820,8 @@ void mmc_perform_opc(struct burn_drive *d)
 {
 	struct command c;
 
-	mmc_function_spy("mmc_perform_opc");
+	if (mmc_function_spy(d, "mmc_perform_opc") <= 0)
+		return;
 
 	scsi_init_command(&c, MMC_SEND_OPC, sizeof(MMC_SEND_OPC));
 /*
@@ -1640,7 +1851,8 @@ int mmc_set_streaming(struct burn_drive *d,
 	char msg[160];
 	unsigned char *pd;
 
-	mmc_function_spy("mmc_set_streaming");
+	if (mmc_function_spy(d, "mmc_set_streaming") <= 0)
+		return 0;
 
 	scsi_init_command(&c, MMC_SET_STREAMING, sizeof(MMC_SET_STREAMING));
 /*
@@ -1722,7 +1934,8 @@ void mmc_set_speed(struct burn_drive *d, int r, int w)
 	int ret, end_lba = 0;
 	struct burn_speed_descriptor *best_sd = NULL;
 
-	mmc_function_spy("mmc_set_speed");
+	if (mmc_function_spy(d, "mmc_set_speed") <= 0)
+		return;
 
 	if (r <= 0 || w <= 0) {
 		/* ts A70712 : now searching for best speed descriptor */
@@ -1842,8 +2055,6 @@ static int mmc_get_configuration_al(struct burn_drive *d, int *alloc_len)
 	d->current_feat21h_link_size = -1;
 	d->current_feat2fh_byte4 = -1;
 
-	mmc_function_spy("mmc_get_configuration");
-
 	scsi_init_command(&c, MMC_GET_CONFIGURATION,
 			 sizeof(MMC_GET_CONFIGURATION));
 /*
@@ -2030,6 +2241,9 @@ void mmc_get_configuration(struct burn_drive *d)
 {
 	int alloc_len = 8, ret;
 
+	if (mmc_function_spy(d, "mmc_get_configuration") <= 0)
+		return;
+
 	/* first command execution to learn Allocation Length */
 	ret = mmc_get_configuration_al(d, &alloc_len);
 /*
@@ -2057,8 +2271,6 @@ static int mmc_read_format_capacities_al(struct burn_drive *d,
 	char msg[160];
 */
 
-	mmc_function_spy("mmc_read_format_capacities");
-
 	if (*alloc_len < 4)
 		return 0;
 
@@ -2195,6 +2407,9 @@ int mmc_read_format_capacities(struct burn_drive *d, int top_wanted)
 {
 	int alloc_len = 4, ret;
 
+	if (mmc_function_spy(d, "mmc_read_format_capacities") <= 0)
+		return 0;
+
 	ret = mmc_read_format_capacities_al(d, &alloc_len, top_wanted);
 /*
 	fprintf(stderr,"LIBBURN_DEBUG: 23h alloc_len = %d , ret = %d\n",
@@ -2211,7 +2426,8 @@ void mmc_sync_cache(struct burn_drive *d)
 {
 	struct command c;
 
-	mmc_function_spy("mmc_sync_cache");
+	if (mmc_function_spy(d, "mmc_sync_cache") <= 0)
+		return;
 
 	scsi_init_command(&c, MMC_SYNC_CACHE, sizeof(MMC_SYNC_CACHE));
 /*
@@ -2219,6 +2435,9 @@ void mmc_sync_cache(struct burn_drive *d)
 	c.oplen = sizeof(MMC_SYNC_CACHE);
 */
 	c.retry = 1;
+
+	c.opcode[1] |= 2; /* ts A70918 : Immed */
+
 	c.page = NULL;
 	c.dir = NO_TRANSFER;
 
@@ -2240,6 +2459,16 @@ void mmc_sync_cache(struct burn_drive *d)
 	}
 
 	d->issue_command(d, &c);
+
+	/* ts A70918 */
+	if (c.error) {
+		d->cancel = 1;
+		return;
+	}
+	if (spc_wait_unit_attention(d, 3600, "SYNCHRONIZE CACHE", 0) <= 0)
+		d->cancel = 1;
+	else
+		d->needs_sync_cache = 0;
 }
 
 
@@ -2253,7 +2482,8 @@ int mmc_read_buffer_capacity(struct burn_drive *d)
 	unsigned char *data;
 	int alloc_len = 12;
 
-	mmc_function_spy("mmc_read_buffer_capacity");
+	if (mmc_function_spy(d, "mmc_read_buffer_capacity") <= 0)
+		return 0;
 
 	scsi_init_command(&c, MMC_READ_BUFFER_CAPACITY,
 			 sizeof(MMC_READ_BUFFER_CAPACITY));
@@ -2319,7 +2549,8 @@ int mmc_format_unit(struct burn_drive *d, off_t size, int flag)
 	char msg[160],descr[80];
 	int full_format_type = 0x00; /* Full Format (or 0x10 for DVD-RW ?) */
 
-	mmc_function_spy("mmc_format_unit");
+	if (mmc_function_spy(d, "mmc_format_unit") <= 0)
+		return 0;
 
 	scsi_init_command(&c, MMC_FORMAT_UNIT, sizeof(MMC_FORMAT_UNIT));
 /*
@@ -2552,8 +2783,6 @@ static int mmc_get_write_performance_al(struct burn_drive *d,
 	/* A61225 : 1 = report about speed descriptors */
 	static int speed_debug = 0;
 
-	mmc_function_spy("mmc_get_write_performance");
-
 	if (d->current_profile <= 0)
 		mmc_get_configuration(d);
 
@@ -2671,6 +2900,9 @@ int mmc_get_write_performance(struct burn_drive *d)
 {
 	int alloc_len = 8, max_descr = 0, ret;
 
+	if (mmc_function_spy(d, "mmc_get_write_performance") <= 0)
+		return 0;
+
 	/* first command execution to learn number of descriptors and 
            dxfer_len */
 	ret = mmc_get_write_performance_al(d, &alloc_len, &max_descr);
@@ -2793,6 +3025,48 @@ int mmc_compose_mode_page_5(struct burn_drive *d,
 }
 
 
+/* A70812 ts */
+int mmc_read_10(struct burn_drive *d, int start,int amount, struct buffer *buf)
+{
+	struct command c;
+
+	if (mmc_function_spy(d, "mmc_read_10") <= 0)
+		return -1;
+;
+	if (amount > BUFFER_SIZE / 2048)
+		return -1;
+
+	scsi_init_command(&c, MMC_READ_10, sizeof(MMC_READ_10));
+	c.dxfer_len = amount * 2048;
+	c.retry = 1;
+	mmc_int_to_four_char(c.opcode + 2, start);
+	c.opcode[7] = (amount >> 8) & 0xFF;
+	c.opcode[8] = amount & 0xFF;
+	c.page = buf;
+	c.page->bytes = 0;
+	c.page->sectors = 0;
+	c.dir = FROM_DRIVE;
+	d->issue_command(d, &c);
+	if (c.error) {
+		char msg[160];
+		sprintf(msg,
+		"SCSI error on read_10(%d,%d): key=%X asc=%2.2Xh ascq=%2.2Xh",
+			start, amount,
+			c.sense[2],c.sense[12],c.sense[13]);
+		if(!d->silent_on_scsi_error)
+			libdax_msgs_submit(libdax_messenger, d->global_index,
+				0x00020144,
+				LIBDAX_MSGS_SEV_SORRY, LIBDAX_MSGS_PRIO_HIGH,
+				msg, 0, 0);
+		return BE_CANCELLED;
+	}
+
+	buf->sectors = amount;
+	buf->bytes = amount * 2048;
+	return 0;
+}
+
+
 /* ts A61021 : the mmc specific part of sg.c:enumerate_common()
 */
 int mmc_setup_drive(struct burn_drive *d)
@@ -2815,6 +3089,7 @@ int mmc_setup_drive(struct burn_drive *d)
 	d->read_buffer_capacity = mmc_read_buffer_capacity;
 	d->format_unit = mmc_format_unit;
 	d->read_format_capacities = mmc_read_format_capacities;
+	d->read_10 = mmc_read_10;
 
 
 	/* ts A70302 */
@@ -2835,6 +3110,7 @@ int mmc_setup_drive(struct burn_drive *d)
 	d->current_feat21h_link_size = -1;
 	d->current_feat2fh_byte4 = -1;
 	d->needs_close_session = 0;
+	d->needs_sync_cache = 0;
 	d->bg_format_status = -1;
 	d->num_format_descr = 0;
 	d->complete_sessions = 0;
