@@ -12,6 +12,9 @@
 #include <stdio.h>
 #include <string.h>
 #include <ctype.h>
+#include <fcntl.h>
+#include <errno.h>
+
 #include "sector.h"
 #include "libburn.h"
 #include "drive.h"
@@ -31,6 +34,12 @@
 #include "read.h"
 #include "options.h"
 
+/* ts A70812 */
+#include "error.h"
+#include "libdax_msgs.h"
+extern struct libdax_msgs *libdax_messenger;
+
+
 void burn_disc_read(struct burn_drive *d, const struct burn_read_opts *o)
 {
 #if 0
@@ -280,3 +289,177 @@ static void flipq(unsigned char *sub)
 	*(sub + 12 + 11) = ~*(sub + 12 + 11);
 }
 */
+
+
+/* ts A70904 */
+/** @param flag bit0=be silent on data shortage */
+int burn_stdio_read(int fd, char *buf, int bufsize, struct burn_drive *d,
+			int flag)
+{
+	int todo, count = 0;
+
+	for(todo = bufsize; todo > 0; ) {
+		count = read(fd, buf + (bufsize - todo), todo);
+		if(count <= 0)
+	break;
+		todo -= count;
+	}
+	if(todo > 0 && !(flag & 1)) {
+		libdax_msgs_submit(libdax_messenger, d->global_index,
+			0x0002014a,
+			LIBDAX_MSGS_SEV_SORRY, LIBDAX_MSGS_PRIO_HIGH,
+			"Cannot read desired amount of data", errno, 0);
+	}
+	if (count < 0)
+		return -1;
+	return (bufsize - todo);
+}
+
+
+/* ts A70812 : API function */
+int burn_read_data(struct burn_drive *d, off_t byte_address,
+                   char data[], off_t data_size, off_t *data_count, int flag)
+{
+	int alignment = 2048, start, upto, chunksize = 1, err, cpy_size, i;
+	int sose_mem = 0, fd = -1, ret;
+	char msg[81], *wpt;
+	struct buffer buf;
+
+	*data_count = 0;
+	sose_mem = d->silent_on_scsi_error;
+
+	if (d->released) {
+		libdax_msgs_submit(libdax_messenger,
+			d->global_index, 0x00020142,
+			LIBDAX_MSGS_SEV_FATAL, LIBDAX_MSGS_PRIO_HIGH,
+			"Drive is not grabbed on random access read", 0, 0);
+		return 0;
+	}
+	if (d->drive_role == 0) {
+		libdax_msgs_submit(libdax_messenger, d->global_index,
+			0x00020146,
+			LIBDAX_MSGS_SEV_FATAL, LIBDAX_MSGS_PRIO_HIGH,
+			"Drive is a virtual placeholder (null-drive)", 0, 0);
+		return 0;
+	} else if (d->drive_role == 3) {
+		libdax_msgs_submit(libdax_messenger, d->global_index,
+			0x00020151,
+			LIBDAX_MSGS_SEV_FATAL, LIBDAX_MSGS_PRIO_HIGH,
+			"Read attempt on write-only drive", 0, 0);
+		return 0;
+	}
+	if ((byte_address % alignment) != 0) {
+		sprintf(msg,
+			"Read start address not properly aligned (%d bytes)",
+			alignment);
+		libdax_msgs_submit(libdax_messenger, d->global_index,
+			0x00020143,
+			LIBDAX_MSGS_SEV_SORRY, LIBDAX_MSGS_PRIO_HIGH,
+			msg, 0, 0);
+		return 0;
+	}
+
+	if (d->busy != BURN_DRIVE_IDLE) {
+		libdax_msgs_submit(libdax_messenger,
+			d->global_index, 0x00020145,
+			LIBDAX_MSGS_SEV_FATAL, LIBDAX_MSGS_PRIO_HIGH,
+			"Drive is busy on attempt to read data", 0, 0);
+		return 0;
+	}
+
+	if (d->drive_role != 1) {
+
+/* <<< We need _LARGEFILE64_SOURCE defined by the build system.
+*/
+#ifndef O_LARGEFILE
+#define O_LARGEFILE 0
+#endif
+
+		fd = open(d->devname, O_RDONLY | O_LARGEFILE);
+		if (fd == -1) {
+			if (errno != ENOENT  || !(flag & 2))
+				libdax_msgs_submit(libdax_messenger,
+				  d->global_index,
+				  0x00020005,
+				  LIBDAX_MSGS_SEV_SORRY, LIBDAX_MSGS_PRIO_HIGH,
+			  "Failed to open device (a pseudo-drive) for reading",
+				   errno, 0);
+			ret = 0; goto ex;
+		}
+		if (lseek(fd, byte_address, SEEK_SET) == -1) {
+			if (!(flag & 2))
+				libdax_msgs_submit(libdax_messenger,
+				  d->global_index,
+				  0x00020147,
+				  LIBDAX_MSGS_SEV_SORRY, LIBDAX_MSGS_PRIO_HIGH,
+				  "Cannot address start byte", errno, 0);
+			ret = 0; goto ex;
+		}
+	}
+
+	d->busy = BURN_DRIVE_READING_SYNC;
+	d->buffer = &buf;
+
+	start = byte_address / 2048;
+	upto = start + data_size / 2048;
+	if (data_size % 2048)
+		upto++;
+	wpt = data;
+	for (; start < upto; start += chunksize) {
+		chunksize = upto - start;
+		if (chunksize > 16) {
+			chunksize = 16;
+			cpy_size = 16 * 2048;
+		} else
+			cpy_size = data_size - *data_count;
+		if (flag & 2)
+			d->silent_on_scsi_error = 1;
+		if (d->drive_role == 1) {
+			err = d->read_10(d, start, chunksize, d->buffer);
+		} else {
+			ret = burn_stdio_read(fd, (char *) d->buffer->data,
+						 cpy_size, d, !!(flag & 2));
+			err = 0;
+			if (ret <= 0)
+				err = BE_CANCELLED;
+		}
+		if (flag & 2) 
+			d->silent_on_scsi_error = sose_mem;
+		if (err == BE_CANCELLED) {
+			/* Try to read a smaller part of the chunk */
+			for (i = 0; i < chunksize - 1; i++) {
+				if (flag & 2)
+					d->silent_on_scsi_error = 1;
+				if (d->drive_role == 1) {
+					err = d->read_10(d, start + i, 1,
+							 d->buffer);
+				} else {
+					ret = burn_stdio_read(fd,
+						(char *) d->buffer->data,
+						2048, d, 1);
+					if (ret <= 0)
+						err = BE_CANCELLED;
+				}
+				if (flag & 2) 
+					d->silent_on_scsi_error = sose_mem;
+				if (err == BE_CANCELLED)
+			break;
+				memcpy(wpt, d->buffer->data, 2048);
+				wpt += 2048;
+				*data_count += 2048;
+			}
+			ret = 0; goto ex;
+		}
+		memcpy(wpt, d->buffer->data, cpy_size);
+		wpt += cpy_size;
+		*data_count += cpy_size;
+	}
+
+	ret = 1;
+ex:;
+	if (fd != -1)
+		close(fd);
+	d->buffer = NULL;
+	d->busy = BURN_DRIVE_IDLE;
+	return ret;
+}
