@@ -48,6 +48,8 @@ sg_issue_command()      sends a SCSI command to the drive, receives reply,
 
 sg_obtain_scsi_adr()    tries to obtain SCSI address parameters.
 
+burn_os_stdio_capacity()  estimates the emulated media space of stdio-drives.
+
 
 Porting hints are marked by the text "PORTING:".
 Send feedback to libburn-hackers@pykix.org .
@@ -73,6 +75,11 @@ Hint: You should also look into sg-freebsd-port.c, which is a younger and
 #include <stdlib.h>
 #include <sys/utsname.h>
 #include <scsi/scsi.h>
+#include <sys/statvfs.h>
+
+/* for ioctl(BLKGETSIZE) */
+#include <linux/fs.h>
+
 
 #include <scsi/sg.h>
 /* Values within sg_io_hdr_t indicating success after ioctl(SG_IO) : */
@@ -83,6 +90,7 @@ Hint: You should also look into sg-freebsd-port.c, which is a younger and
 
 
 /* ts A61211 : to eventually recognize CD devices on /dev/sr* */
+#include <limits.h>
 #include <linux/cdrom.h>
 
 
@@ -194,8 +202,9 @@ extern int burn_sg_open_abort_busy;
 
 
 /* ts A60821
-   <<< debug: for tracing calls which might use open drive fds */
-int mmc_function_spy(char * text);
+   debug: for tracing calls which might use open drive fds
+          or for catching SCSI usage of emulated drives. */
+int mmc_function_spy(struct burn_drive *d, char * text);
 
 
 /* ------------------------------------------------------------------------ */
@@ -932,7 +941,8 @@ int sg_grab(struct burn_drive *d)
 
 /* ts A60821
    <<< debug: for tracing calls which might use open drive fds */
-	mmc_function_spy("sg_grab");
+	if (mmc_function_spy(d, "sg_grab") <= 0)
+		return 0;
 
 
 	/* ts A60813 - A60927
@@ -960,7 +970,7 @@ int sg_grab(struct burn_drive *d)
 
 		/* ts A60821
    		<<< debug: for tracing calls which might use open drive fds */
-		mmc_function_spy("sg_grab ----------- opening");
+		mmc_function_spy(NULL, "sg_grab ----------- opening");
 
 		/* ts A70409 : DDLP-B */
 		/* >>> obtain single lock on d->devname */
@@ -1018,7 +1028,8 @@ int sg_release(struct burn_drive *d)
 {
 	/* ts A60821
    	<<< debug: for tracing calls which might use open drive fds */
-	mmc_function_spy("sg_release");
+	if (mmc_function_spy(d, "sg_release") <= 0)
+		return 0;
 
 	if (d->fd < 1) {
 		burn_print(1, "release an ungrabbed drive.  die\n");
@@ -1027,7 +1038,7 @@ int sg_release(struct burn_drive *d)
 
 	/* ts A60821
    	<<< debug: for tracing calls which might use open drive fds */
-	mmc_function_spy("sg_release ----------- closing");
+	mmc_function_spy(NULL, "sg_release ----------- closing");
 
 	sg_close_drive(d);
 	return 0;
@@ -1124,7 +1135,7 @@ int sg_issue_command(struct burn_drive *d, struct command *c)
 	char buf[161];
 	sprintf(buf,"sg_issue_command   d->fd= %d  d->released= %d\n",
 		d->fd,d->released);
-	mmc_function_spy(buf);
+	mmc_function_spy(NULL, buf);
 
 #ifdef Libburn_log_sg_commandS
 	/* ts A61030 */
@@ -1352,3 +1363,60 @@ int sg_is_enumerable_adr(char *adr)
 }
 
 
+/* ts A70909 */
+/** Estimate the potential payload capacity of a file address.
+    @param path  The address of the file to be examined. If it does not
+                 exist yet, then the directory will be inquired.
+    @param bytes The pointed value gets modified, but only if an estimation is
+                 possible.
+    @return      -2 = cannot perform necessary operations on file object
+                 -1 = neither path nor dirname of path exist
+                  0 = could not estimate size capacity of file object
+                  1 = estimation has been made, bytes was set
+*/
+int burn_os_stdio_capacity(char *path, off_t *bytes)
+{
+	struct stat stbuf;
+	struct statvfs vfsbuf;
+	char testpath[4096], *cpt;
+	long blocks;
+	int open_mode = O_RDONLY, fd, ret;
+	off_t add_size = 0;
+
+	testpath[0] = 0;
+	blocks = *bytes / 512;
+	if (stat(path, &stbuf) == -1) {
+		strcpy(testpath, path);
+		cpt = strrchr(testpath, '/');
+		if(cpt == NULL)
+			strcpy(testpath, ".");
+		else if(cpt == testpath)
+			testpath[1] = 0;
+		else
+			*cpt = 0;
+		if (stat(testpath, &stbuf) == -1)
+			return -1;
+	} else if(S_ISBLK(stbuf.st_mode)) {
+		fd = open(path, open_mode);
+		if (fd == -1)
+			return -2;
+		ret = ioctl(fd, BLKGETSIZE, &blocks);
+		close(fd);
+		if (ret == -1)
+			return -2;
+		*bytes = ((off_t) blocks) * (off_t) 512;
+	} else if(S_ISREG(stbuf.st_mode)) {
+		add_size = stbuf.st_blocks * (off_t) 512;
+		strcpy(testpath, path);
+	} else
+		return 0;
+
+	if (testpath[0]) {	
+		if (statvfs(testpath, &vfsbuf) == -1)
+			return -2;
+		*bytes = add_size + ((off_t) vfsbuf.f_bsize) *
+						(off_t) vfsbuf.f_bavail;
+	}
+	return 1;
+}
+
