@@ -27,6 +27,11 @@
 #include "libdax_msgs.h"
 extern struct libdax_msgs *libdax_messenger;
 
+/* ts A70910
+   debug: for tracing calls which might use open drive fds
+          or for catching SCSI usage of emulated drives. */
+int mmc_function_spy(struct burn_drive *d, char * text);
+
 
 /* spc command set */
 /* ts A70519 : allocation length byte 3+4 was 0,255 */
@@ -63,6 +68,9 @@ int spc_test_unit_ready_r(struct burn_drive *d, int *key, int *asc, int *ascq)
 {
 	struct command c;
 
+	if (mmc_function_spy(d, "test_unit_ready") <= 0)
+		return 0;
+
 	scsi_init_command(&c, SPC_TEST_UNIT_READY,sizeof(SPC_TEST_UNIT_READY));
 /*
 	c.oplen = sizeof(SPC_TEST_UNIT_READY);
@@ -90,19 +98,73 @@ int spc_test_unit_ready(struct burn_drive *d)
 
 
 /* ts A70315 */
+/** @param flag bit0=do not wait 0.1 seconds before first test unit ready */
 /** Wait until the drive state becomes clear or until max_usec elapsed */
-int spc_wait_unit_attention(struct burn_drive *d, int max_sec)
+int spc_wait_unit_attention(struct burn_drive *d, int max_sec, char *cmd_text,
+				int flag)
 {
-	int i, ret, key, asc, ascq;
+	int i, ret = 1, key = 0, asc = 0, ascq = 0;
+	char msg[160];
 
-	for(i=0; i < max_sec; i++) {
+	if (!(flag & 1))
+		usleep(100000);
+	for(i = !(flag & 1); i < max_sec * 10; i++) {
 		ret = spc_test_unit_ready_r(d, &key, &asc, &ascq);
-		if(ret > 0 || key!=0x2 || asc!=0x4) /* ready or error */
+
+/* <<< 
+		fprintf(stderr,
+"libburn_EXPERIMENTAL: i= %d  ret= %d  key= %X  asc= %2.2X  ascq= %2.2X\n",
+		i, ret, (unsigned) key, (unsigned) asc, (unsigned) ascq);
+*/
+
+		if(ret > 0) /* ready */
 	break;
-		usleep(1000000);
+		if(key!=0x2 || asc!=0x4) {
+			if (key == 0x2 && asc == 0x3A) {
+				ret = 1; /* medium not present = ok */
+/* <<<
+  ts A70912 :
+  My LG GSA-4082B on asynchronous load:
+    first it reports no media 2,3A,00,
+    then it reports not ready 2,04,00,
+    further media inquiry retrieves wrong data
+
+				if(i<=100)
+					goto slumber;
+*/
+	break;
+			}
+			if (key == 0x6 && asc == 0x28 && ascq == 0x00)
+				/* media change notice = try again */
+				goto slumber;
+
+			sprintf(msg,
+		"Asynchromous SCSI error on %s: key=%X asc=%2.2Xh ascq=%2.2Xh",
+			 	cmd_text, (unsigned) key, (unsigned) asc,
+				(unsigned) ascq);
+			libdax_msgs_submit(libdax_messenger, d->global_index,
+				0x0002014d,
+				LIBDAX_MSGS_SEV_SORRY, LIBDAX_MSGS_PRIO_HIGH,
+				msg, 0, 0);
+			d->cancel = 1;
+	break;
+		}
+slumber:;
+		usleep(100000);
 	}
-	if (i < max_sec)
+
+	sprintf(msg, "Async %s %s after %d.%d seconds",
+		cmd_text, (ret > 0 ? "succeeded" : "failed"), i / 10, i % 10);
+	libdax_msgs_submit(libdax_messenger, d->global_index, 0x00020150,
+		LIBDAX_MSGS_SEV_DEBUG, LIBDAX_MSGS_PRIO_LOW, msg, 0, 0);
+
+	if (i < max_sec * 10)
 		return (ret > 0);
+
+	sprintf(msg, "Timeout (%d s) with asynchronous SCSI command %s\n",
+	 	max_sec, cmd_text);
+	libdax_msgs_submit(libdax_messenger, d->global_index, 0x0002014f,
+		LIBDAX_MSGS_SEV_SORRY, LIBDAX_MSGS_PRIO_HIGH, msg, 0, 0);
 	return 0;
 }
 
@@ -111,6 +173,9 @@ void spc_request_sense(struct burn_drive *d, struct buffer *buf)
 {
 	struct command c;
 
+	if (mmc_function_spy(d, "request_sense") <= 0)
+		return;
+
 	scsi_init_command(&c, SPC_REQUEST_SENSE, sizeof(SPC_REQUEST_SENSE));
 	c.retry = 0;
 /*
@@ -130,6 +195,9 @@ int spc_get_erase_progress(struct burn_drive *d)
 {
 	struct buffer b;
 
+	if (mmc_function_spy(d, "get_erase_progress") <= 0)
+		return 0;
+
 	spc_request_sense(d, &b);
 	return (b.data[16] << 8) | b.data[17];
 }
@@ -140,6 +208,9 @@ void spc_inquiry(struct burn_drive *d)
 	struct burn_scsi_inquiry_data *id;
 	struct command c;
 
+	if (mmc_function_spy(d, "inquiry") <= 0)
+		return;
+
 	scsi_init_command(&c, SPC_INQUIRY, sizeof(SPC_INQUIRY));
 /*
 	memcpy(c.opcode, SPC_INQUIRY, sizeof(SPC_INQUIRY));
@@ -171,6 +242,9 @@ void spc_prevent(struct burn_drive *d)
 {
 	struct command c;
 
+	if (mmc_function_spy(d, "prevent") <= 0)
+		return;
+
 	scsi_init_command(&c, SPC_PREVENT, sizeof(SPC_PREVENT));
 /*
 	memcpy(c.opcode, SPC_PREVENT, sizeof(SPC_PREVENT));
@@ -186,6 +260,9 @@ void spc_allow(struct burn_drive *d)
 {
 	struct command c;
 
+	if (mmc_function_spy(d, "allow") <= 0)
+		return;
+
 	scsi_init_command(&c, SPC_ALLOW, sizeof(SPC_ALLOW));
 /*
 	memcpy(c.opcode, SPC_ALLOW, sizeof(SPC_ALLOW));
@@ -368,6 +445,9 @@ void spc_sense_caps(struct burn_drive *d)
 {
 	int alloc_len, start_len = 22, ret;
 
+	if (mmc_function_spy(d, "sense_caps") <= 0)
+		return;
+
 	/* first command execution to learn Allocation Length */
 	alloc_len = start_len;
 	ret = spc_sense_caps_al(d, &alloc_len, 1);
@@ -389,6 +469,9 @@ void spc_sense_error_params(struct burn_drive *d)
 	unsigned char *page;
 	struct command c;
 
+	if (mmc_function_spy(d, "sense_error_params") <= 0)
+		return;
+
 	scsi_init_command(&c, SPC_MODE_SENSE, sizeof(SPC_MODE_SENSE));
 /*
 	memcpy(c.opcode, SPC_MODE_SENSE, sizeof(SPC_MODE_SENSE));
@@ -419,6 +502,9 @@ void spc_select_error_params(struct burn_drive *d,
 	struct buffer buf;
 	struct command c;
 
+	if (mmc_function_spy(d, "select_error_params") <= 0)
+		return;
+
 	scsi_init_command(&c, SPC_MODE_SELECT, sizeof(SPC_MODE_SELECT));
 /*
 	memcpy(c.opcode, SPC_MODE_SELECT, sizeof(SPC_MODE_SELECT));
@@ -457,6 +543,9 @@ void spc_sense_write_params(struct burn_drive *d)
 	unsigned char *page;
 	struct command c;
 
+	if (mmc_function_spy(d, "sense_write_params") <= 0)
+		return;
+
 	/* ts A61007 : Done in soft at only caller burn_drive_grab() */
 	/* a ssert(d->mdata->cdr_write || d->mdata->cdrw_write ||
 	       d->mdata->dvdr_write || d->mdata->dvdram_write); */
@@ -477,12 +566,16 @@ void spc_sense_write_params(struct burn_drive *d)
 	c.dir = FROM_DRIVE;
 	d->issue_command(d, &c);
 
-	size = c.page->data[0] * 256 + c.page->data[1];
+	/* ts A71128 : do not interpret reply if error */
 	m = d->mdata;
-	page = c.page->data + 8;
-	burn_print(1, "write page length 0x%x\n", page[1]);
-	m->write_page_length = page[1];
-	m->write_page_valid = 1;
+	if(!c.error) {
+		size = c.page->data[0] * 256 + c.page->data[1];
+		page = c.page->data + 8;
+		burn_print(1, "write page length 0x%x\n", page[1]);
+		m->write_page_length = page[1];
+		m->write_page_valid = 1;
+	} else
+		m->write_page_valid = 0;
 	mmc_read_disc_info(d);
 
 	/* ts A70212 : try to setup d->media_capacity_remaining */
@@ -514,6 +607,9 @@ void spc_select_write_params(struct burn_drive *d,
 	struct buffer buf;
 	struct command c;
 
+	if (mmc_function_spy(d, "select_write_params") <= 0)
+		return;
+
 	/* ts A61007 : All current callers are safe. */
 	/* a ssert(o->drive == d); */
 
@@ -554,6 +650,9 @@ void spc_select_write_params(struct burn_drive *d,
 
 void spc_getcaps(struct burn_drive *d)
 {
+	if (mmc_function_spy(d, "getcaps") <= 0)
+		return;
+
 	spc_inquiry(d);
 	spc_sense_caps(d);
 	spc_sense_error_params(d);
@@ -573,6 +672,9 @@ void spc_probe_write_modes(struct burn_drive *d)
 	int last_try = 0;
 	struct command c;
 
+	if (mmc_function_spy(d, "spc_probe_write_modes") <= 0)
+		return;
+
 	/* ts A70213 : added pseudo try_write_type 4 to set a suitable mode */
 	while (try_write_type != 5) {
 		burn_print(9, "trying %d, %d\n", try_write_type,
@@ -739,11 +841,8 @@ int burn_scsi_setup_drive(struct burn_drive *d, int bus_no, int host_no,
 	d->silent_on_scsi_error = 0;
 
 
-	d->idata = malloc(sizeof(struct burn_scsi_inquiry_data));
-	d->idata->valid = 0;
-	d->mdata = malloc(sizeof(struct scsi_mode_data));
-	d->mdata->valid = 0;
-	d->mdata->speed_descriptors = NULL;
+	d->idata = calloc(1, sizeof(struct burn_scsi_inquiry_data));
+	d->mdata = calloc(1, sizeof(struct scsi_mode_data));
 
 	/* ts A61007 : obsolete Assert in drive_getcaps() */
 	if(d->idata == NULL || d->mdata == NULL) {
@@ -752,6 +851,9 @@ int burn_scsi_setup_drive(struct burn_drive *d, int bus_no, int host_no,
 	                "Could not allocate new drive object", 0, 0);
 		return -1;
 	}
+	d->idata->valid = 0;
+	d->mdata->valid = 0;
+	d->mdata->speed_descriptors = NULL;
 	if(!(flag & 1)) {
 		ret = spc_setup_drive(d);
 		if (ret<=0)
