@@ -18,11 +18,16 @@
 #define Libburn_sao_can_appenD 1
 */
 
-
+#include <sys/types.h>
 #include <stdio.h>
 #include <string.h>
 #include <ctype.h>
 #include <stdlib.h>
+#include <fcntl.h>
+#include <errno.h>
+#include <sys/stat.h>
+#include <sys/time.h>
+
 #include "error.h"
 #include "sector.h"
 #include "libburn.h"
@@ -38,6 +43,7 @@
 #include "write.h"
 #include "options.h"
 #include "structure.h"
+#include "source.h"
 
 #include "libdax_msgs.h"
 extern struct libdax_msgs *libdax_messenger;
@@ -112,7 +118,9 @@ void type_to_form(int mode, unsigned char *ctladr, int *form)
 		*form |= 0x40;
 }
 
-int burn_write_flush(struct burn_write_opts *o, struct burn_track *track)
+
+/* ts A71002 : outsourced from burn_write_flush() : no sync cache here */
+int burn_write_flush_buffer(struct burn_write_opts *o,struct burn_track *track)
 {
 	struct burn_drive *d = o->drive;
 
@@ -133,13 +141,25 @@ int burn_write_flush(struct burn_write_opts *o, struct burn_track *track)
 		d->buffer->bytes = 0;
 		d->buffer->sectors = 0;
 	}
+	return 1;
+}
+
+
+int burn_write_flush(struct burn_write_opts *o, struct burn_track *track)
+{
+	int ret;
+	struct burn_drive *d = o->drive;
+
+	ret = burn_write_flush_buffer(o, track);
+	if (ret <= 0)
+		return ret;
 	d->sync_cache(d);
 	return 1;
 }
 
 
-/* ts A61030 */
-int burn_write_close_track(struct burn_write_opts *o, struct burn_session *s,
+/* ts A71002 : outsourced from burn_write_close_track() */
+int burn_write_track_minsize(struct burn_write_opts *o, struct burn_session *s,
 				int tnum)
 {
 	char msg[81];
@@ -147,11 +167,6 @@ int burn_write_close_track(struct burn_write_opts *o, struct burn_session *s,
 	struct burn_track *t;
 	int todo, step, cancelled, seclen;
 
-	/* ts A61106 */
-#ifdef Libburn_experimental_no_close_tracK
-	return 1;
-#endif
-
 	d = o->drive;
 	t = s->track[tnum];
 
@@ -185,8 +200,26 @@ int burn_write_close_track(struct burn_write_opts *o, struct burn_session *s,
 		}
 		d->cancel = cancelled;
 	}
+	return 1;
+}
+
+
+/* ts A61030 */
+int burn_write_close_track(struct burn_write_opts *o, struct burn_session *s,
+				int tnum)
+{
+	char msg[81];
+	struct burn_drive *d;
+	struct burn_track *t;
+
+	/* ts A61106 */
+#ifdef Libburn_experimental_no_close_tracK
+	return 1;
+#endif
+
+	d = o->drive;
+	t = s->track[tnum];
 
-	/* ts A61102 */
 	d->busy = BURN_DRIVE_CLOSING_TRACK;
 
 	sprintf(msg, "Closing track %2.2d", tnum+1);
@@ -201,13 +234,13 @@ int burn_write_close_track(struct burn_write_opts *o, struct burn_session *s,
 	*/
 	d->close_track_session(o->drive, 0, 0xff);
 
-	/* ts A61102 */
 	d->busy = BURN_DRIVE_WRITING;
 
 	return 1;
 }
 
 
+
 /* ts A61030 */
 int burn_write_close_session(struct burn_write_opts *o, struct burn_session *s)
 {
@@ -804,14 +837,18 @@ int burn_write_track(struct burn_write_opts *o, struct burn_session *s,
 	/* ts A61103 */
 	ret = 1;
 ex:;
+	if (d->cancel)
+		burn_source_cancel(t->source);
 	if (o->write_type == BURN_WRITE_TAO) {
 
-		/* ts A61103 */
-		/* >>> if cancelled: ensure that at least 600 kB get written */
-
-		if (!burn_write_flush(o, t))
+		/* ts A71002 */
+		if (!burn_write_flush_buffer(o, t))
 			ret = 0;
 
+		/* Ensure that at least 600 kB get written */
+		burn_write_track_minsize(o, s, tnum);
+		d->sync_cache(d);
+
 		/* ts A61030 */
 		if (burn_write_close_track(o, s, tnum) <= 0)
 			ret = 0;
@@ -910,6 +947,13 @@ int burn_precheck_write(struct burn_write_opts *o, struct burn_disc *disc,
 	reason_pt= reasons;
 	reasons[0] = 0;
 
+	if (d->drive_role == 0) {
+		sprintf(reasons,
+			 "DRIVE: is a virtual placeholder (null-drive)");
+		no_media = 1;
+		goto ex;
+	}
+
 	/* check write mode against write job */
 	wt = burn_write_opts_auto_write_type(o, disc, reasons, 1);
 	if (wt == BURN_WRITE_NONE) {
@@ -922,25 +966,28 @@ int burn_precheck_write(struct burn_write_opts *o, struct burn_disc *disc,
 	reason_pt= reasons + strlen(reasons);
 	if (d->status == BURN_DISC_UNSUITABLE)
 		goto unsuitable_profile;
-	if (d->current_profile == 0x09 || d->current_profile == 0x0a) {
+	if (d->drive_role == 2 ||
+		d->current_profile == 0x1a || d->current_profile == 0x12) { 
+		/* DVD+RW , DVD-RAM , emulated drive on stdio file */
+		if (o->start_byte >= 0 && (o->start_byte % 2048))
+			strcat(reasons,
+			 "write start address not properly aligned to 2048, ");
+	} else if (d->current_profile == 0x09 || d->current_profile == 0x0a) {
+		/* CD-R , CD-RW */
 		if (!burn_disc_write_is_ok(o, disc, (!!silent) << 1))
 			strcat(reasons, "unsuitable track mode found, ");
 		if (o->start_byte >= 0)
 			strcat(reasons, "write start address not supported, ");
-	} else if (d->current_profile == 0x1a || d->current_profile == 0x12) { 
-		/* DVD+RW , DVD-RAM */
-		if (o->start_byte >= 0 && (o->start_byte % 2048))
-			strcat(reasons,
-			 "write start address not properly aligned to 2048, ");
 	} else if (d->current_profile == 0x13) {
 		/* DVD-RW Restricted Overwrite */
 		if (o->start_byte >= 0 && (o->start_byte % 32768))
 			strcat(reasons,
 			  "write start address not properly aligned to 32k, ");
-	} else if (d->current_profile == 0x11 || d->current_profile == 0x14 ||
+	} else if (d->drive_role == 3 ||
+		   d->current_profile == 0x11 || d->current_profile == 0x14 ||
 	           d->current_profile == 0x15 ||
 	           d->current_profile == 0x1b || d->current_profile == 0x2b ) {
-		/* DVD-R* Sequential , DVD+R[/DL] */
+		/* DVD-R* Sequential , DVD+R[/DL] , sequential stdio "drive" */
 		if (o->start_byte >= 0)
 			strcat(reasons, "write start address not supported, ");
 	} else {
@@ -1177,7 +1224,7 @@ int burn_dvd_write_track(struct burn_write_opts *o,
 	}
 
 
-	/* >>> ts A70215 : what about offset padding ? */
+	/* (offset padding is done within sector_data()) */
 
 	burn_disc_init_track_status(o, s, tnum, sectors);
 	for (i = 0; open_ended || i < sectors; i++) {
@@ -1200,7 +1247,7 @@ int burn_dvd_write_track(struct burn_write_opts *o,
 		d->progress.sector++;
 	}
 	
-	/* >>> ts A70215 : what about tail padding ? */
+	/* (tail padding is done in sector_data()) */
 
 	/* Pad up buffer to next full o->obs (usually 32 kB) */
 	if (o->obs_pad && out->bytes > 0 && out->bytes < o->obs) {
@@ -1229,6 +1276,8 @@ int burn_dvd_write_track(struct burn_write_opts *o,
 	}
 	ret = 1;
 ex:;
+	if (d->cancel)
+		burn_source_cancel(t->source);
 	if (!is_flushed)
 		d->sync_cache(d); /* burn_write_flush() was not called */
 	return ret;
@@ -1431,8 +1480,6 @@ int burn_disc_setup_dvd_minus_rw(struct burn_write_opts *o,
 
 	/* >>> perform OPC if needed */;
 
-	/* >>> */;
-
 	return 1;
 }
 
@@ -1473,63 +1520,10 @@ int burn_dvd_write_sync(struct burn_write_opts *o,
 	struct burn_track *t;
 	char msg[160];
 
-#ifndef Libburn_precheck_write_ruleS
-	int exotic_track = 0, dao_is_ok, sx, tx, mode;
-#endif
-
 	d->needs_close_session = 0;
 
-#ifndef Libburn_precheck_write_ruleS
-	/* <<< covered by burn_precheck_write() */
-	for (sx = 0; sx < disc->sessions; sx++)
-		for (tx = 0 ; tx < disc->session[sx]->tracks; tx++) {
-			mode = disc->session[sx]->track[tx]->mode;
-			if (disc->session[sx]->track[tx]->mode != BURN_MODE1)
-				exotic_track = 1;
-		}
-	if (exotic_track) {
-		sprintf(msg,"DVD Media are unsuitable for desired track type");
-		libdax_msgs_submit(libdax_messenger, d->global_index,
-			0x00020123,
-			LIBDAX_MSGS_SEV_SORRY, LIBDAX_MSGS_PRIO_HIGH,
-			msg, 0, 0);
-		goto early_failure;
-	}
-
-	/* <<< covered by burn_precheck_write() */
-	if (d->current_profile == 0x1a || d->current_profile == 0x13 ||
-	    d->current_profile == 0x12) {
-		 /* DVD+RW , DVD-RW Restricted Overwrite , DVD-RAM */
-		if (disc->sessions!=1 || disc->session[0]->tracks>1
-			|| o->multi ) {
-			sprintf(msg,
-	"Burning is restricted to a single track and no multi-session on %s",
-				d->current_profile_text);
-			libdax_msgs_submit(libdax_messenger, d->global_index,
-				0x0002011f,
-				LIBDAX_MSGS_SEV_SORRY, LIBDAX_MSGS_PRIO_HIGH,
-				msg, 0, 0);
-			goto early_failure;
-		}
-	}
-#endif /* ! Libburn_precheck_write_ruleS */
-
 	if (d->current_profile == 0x1a || d->current_profile == 0x12) { 
 		/* DVD+RW , DVD-RAM */
-
-#ifndef Libburn_precheck_write_ruleS
-		/* <<< covered by burn_precheck_write() */
-		if (o->start_byte >= 0 && (o->start_byte % 2048)) {
-			sprintf(msg,
-			  "Write start address not properly aligned to 2048");
-			libdax_msgs_submit(libdax_messenger, d->global_index,
-				0x00020125,
-				LIBDAX_MSGS_SEV_SORRY, LIBDAX_MSGS_PRIO_HIGH,
-				msg, 0, 0);
-			goto early_failure;
-		}
-#endif /* ! Libburn_precheck_write_ruleS */
-
 		ret = 1;
 		if (d->current_profile == 0x1a)
 			ret = burn_disc_setup_dvd_plus_rw(o, disc);
@@ -1546,20 +1540,6 @@ int burn_dvd_write_sync(struct burn_write_opts *o,
 
 	} else if (d->current_profile == 0x13) {
 		 /* DVD-RW Restricted Overwrite */
-
-#ifndef Libburn_precheck_write_ruleS
-		/* <<< covered by burn_precheck_write() */
-		if (o->start_byte >= 0 && (o->start_byte % 32768)) {
-			sprintf(msg,
-			  "Write start address not properly aligned to 32K");
-			libdax_msgs_submit(libdax_messenger, d->global_index,
-				0x00020125,
-				LIBDAX_MSGS_SEV_SORRY, LIBDAX_MSGS_PRIO_HIGH,
-				msg, 0, 0);
-			goto early_failure;
-		}
-#endif /* ! Libburn_precheck_write_ruleS */
-
 		ret = burn_disc_setup_dvd_minus_rw(o, disc);
 		if (ret <= 0) {
 			sprintf(msg,
@@ -1580,69 +1560,6 @@ int burn_dvd_write_sync(struct burn_write_opts *o,
 		t = disc->session[0]->track[0];
 		o_end = ( burn_track_is_open_ended(t) && !o->fill_up_media );
 		default_size = burn_track_get_default_size(t);
-
-#ifndef Libburn_precheck_write_ruleS
-		dao_is_ok =
-			(disc->sessions == 1 &&
-			 disc->session[0]->tracks == 1 &&
-			 (default_size > 0 || !o_end) &&
-			 (!o->multi) && d->status == BURN_DISC_BLANK
-			);
-		if (o->write_type == BURN_WRITE_TAO && 
-		    !d->current_has_feat21h) {
-
-			/* <<< ??? keep this automatic write type change ? */
-			if (dao_is_ok) {
-				o->write_type = BURN_WRITE_SAO;
-				libdax_msgs_submit(libdax_messenger,
-				  d->global_index, 0x00020134,
-				  LIBDAX_MSGS_SEV_NOTE, LIBDAX_MSGS_PRIO_HIGH,
-				  "Defaulted TAO to DAO (lack of feature 21h)",
-				  0, 0);
-
-			/* <<< covered by burn_precheck_write() */
-			} else {
-				libdax_msgs_submit(libdax_messenger,
-				  d->global_index, 0x00020135,
-				  LIBDAX_MSGS_SEV_SORRY, LIBDAX_MSGS_PRIO_HIGH,
-	"Cannot perform TAO (lack of feature 21h), job unsuitable for DAO",
-				  0, 0);
-				goto early_failure;
-			}
-
-		} else if (o->write_type == BURN_WRITE_SAO && !dao_is_ok) {
-
-			/* <<< covered by burn_precheck_write() */
-			libdax_msgs_submit(libdax_messenger, d->global_index,
-				0x00020136,
-				LIBDAX_MSGS_SEV_SORRY, LIBDAX_MSGS_PRIO_HIGH,
-				"DAO burning is restricted to a single fixed size track and no multi-session",
-				0, 0);
-
-			/* <<< ??? keep this automatic advise ? */
-			if (d->current_has_feat21h)
-				libdax_msgs_submit(libdax_messenger,
-					d->global_index, 0x00020137,
-					LIBDAX_MSGS_SEV_HINT,
-					LIBDAX_MSGS_PRIO_HIGH,
-				"TAO would be possible and could do the job",
-					0, 0);
-
-			goto early_failure;
-
-		}
-
-		/* <<< covered by burn_precheck_write() */
-		if (o->start_byte >= 0) {
-			sprintf(msg, "Write start address not supported");
-			libdax_msgs_submit(libdax_messenger, d->global_index,
-				0x00020124,
-				LIBDAX_MSGS_SEV_SORRY, LIBDAX_MSGS_PRIO_HIGH,
-				msg, 0, 0);
-			goto early_failure;
-		}
-#endif /* ! Libburn_precheck_write_ruleS */
-
 		if (o->write_type == BURN_WRITE_SAO && o_end) {
 			sprintf(msg, "Activated track default size %.f",
 				(double) default_size);
@@ -1670,11 +1587,6 @@ int burn_dvd_write_sync(struct burn_write_opts *o,
 		t = disc->session[0]->track[0];
 		o_end = ( burn_track_is_open_ended(t) && !o->fill_up_media );
 		default_size = burn_track_get_default_size(t);
-
-#ifndef Libburn_precheck_write_ruleS
-		/* >>> oldfashioned checks hopefully never re-enabled */
-#endif /* ! Libburn_precheck_write_ruleS */
-
 		if (o->write_type == BURN_WRITE_SAO && o_end) {
 			sprintf(msg, "Activated track default size %.f",
 				(double) default_size);
@@ -1696,19 +1608,6 @@ int burn_dvd_write_sync(struct burn_write_opts *o,
 		}
 		/* ??? padding needed ??? cowardly doing it for now */
 		o->obs_pad = 1; /* fill-up track's last 32k buffer */
-
-#ifndef Libburn_precheck_write_ruleS
-	/* <<< covered by burn_precheck_write() */
-	} else {
-		sprintf(msg, "Unsuitable media detected. Profile %4.4Xh  %s",
-			d->current_profile, d->current_profile_text);
-		libdax_msgs_submit(libdax_messenger, d->global_index,
-			0x0002011e,
-			LIBDAX_MSGS_SEV_SORRY, LIBDAX_MSGS_PRIO_HIGH,
-			msg, 0, 0);
-		goto early_failure;
-#endif /* ! Libburn_precheck_write_ruleS */
-
 	}
 	o->obs = 32*1024; /* buffer flush trigger for sector.c:get_sector() */
 
@@ -1743,6 +1642,323 @@ early_failure:;
 }
 
 
+/* ts A70904 */
+int burn_stdio_open_write(struct burn_drive *d, off_t start_byte,
+			 int sector_size, int flag)
+{
+
+/* <<< We need _LARGEFILE64_SOURCE defined by the build system.
+*/
+#ifndef O_LARGEFILE
+#define O_LARGEFILE 0
+#endif
+
+	int fd = -1;
+	int mode = O_RDWR | O_CREAT | O_LARGEFILE;
+	char msg[160];
+
+	if (d->devname[0] == 0) /* null drives should not come here */
+		return -1;
+	fd = burn_drive__fd_from_special_adr(d->devname);
+	if (fd >= 0)
+		fd = dup(fd); /* check validity and make closeable */
+	else
+		fd = open(d->devname, mode, S_IRUSR | S_IWUSR);
+	if (fd == -1) {
+		libdax_msgs_submit(libdax_messenger, d->global_index,
+			0x00020005,
+			LIBDAX_MSGS_SEV_SORRY, LIBDAX_MSGS_PRIO_HIGH,
+			"Failed to open device (a pseudo-drive)", errno, 0);
+		d->cancel = 1;
+		return -1;
+	} 
+	if (start_byte < 0)
+		start_byte = 0;
+	if (d->drive_role == 2)
+		if (lseek(fd, start_byte, SEEK_SET)==-1) {
+			sprintf(msg, "Cannot address start byte %.f",
+			 	(double) start_byte);
+			libdax_msgs_submit(libdax_messenger, d->global_index,
+				0x00020147,
+				LIBDAX_MSGS_SEV_SORRY, LIBDAX_MSGS_PRIO_HIGH,
+				msg, errno, 0);
+			close(fd);
+			d->cancel = 1;
+			fd = -1;
+		}
+	d->nwa = start_byte / sector_size;
+	return fd;
+}
+
+
+/* ts A70904 */
+int burn_stdio_read_source(struct burn_source *source, char *buf, int bufsize,
+			 	struct burn_write_opts *o, int flag)
+{
+	int count= 0, todo;
+
+	for(todo = bufsize; todo > 0; todo -= count) {
+		if(source->read!=NULL)
+			count = source->read(source,
+			    (unsigned char *) (buf + (bufsize - todo)), todo);
+		else
+			count = source->read_xt(source,
+			    (unsigned char *) (buf + (bufsize - todo)), todo);
+		if (count <= 0)
+	break;
+	}
+	return (bufsize - todo);
+}
+
+
+/* ts A70904 */
+int burn_stdio_write(int fd, char *buf, int count, struct burn_drive *d, 
+			 int flag)
+{
+	if (write(fd, buf, count) != count) {
+		libdax_msgs_submit(libdax_messenger, d->global_index,
+			0x00020148,
+			LIBDAX_MSGS_SEV_SORRY, LIBDAX_MSGS_PRIO_HIGH,
+			"Cannot write desired amount of data", errno, 0);
+		d->cancel = 1;
+		return 0;
+	}
+	return count;
+}
+
+
+/* ts A70910 : to be used as burn_drive.write(), emulating mmc_write() */
+int burn_stdio_mmc_write(struct burn_drive *d, int start, struct buffer *buf)
+{
+	int ret;
+	off_t start_byte;
+
+	if (d->cancel)
+		return BE_CANCELLED;
+	if (d->stdio_fd < 0) {
+
+		/* >>> program error */;
+
+		d->cancel = 1;
+		return BE_CANCELLED;
+	}
+	if (start != d->nwa) {
+		char msg[80];
+
+		start_byte = ((off_t) start) * 
+				(off_t) (buf->bytes / buf->sectors);
+		if (lseek(d->stdio_fd, start_byte, SEEK_SET)==-1) {
+			sprintf(msg, "Cannot address start byte %.f",
+			 	(double) start_byte);
+			libdax_msgs_submit(libdax_messenger, d->global_index,
+				0x00020147,
+				LIBDAX_MSGS_SEV_SORRY, LIBDAX_MSGS_PRIO_HIGH,
+				msg, errno, 0);
+			d->cancel = 1;
+			return BE_CANCELLED;
+		}
+		d->nwa = start;
+	}
+	ret = burn_stdio_write(d->stdio_fd,(char *)buf->data, buf->bytes, d,0);
+	if (ret <= 0)
+		return BE_CANCELLED;
+	d->nwa += buf->sectors;
+	return 0;
+}
+
+
+/* ts A70910 : to be used as burn_drive.write(),
+               emulating mmc_write() with simulated writing. */
+int burn_stdio_mmc_dummy_write(struct burn_drive *d, int start,
+							struct buffer *buf)
+{
+	if (d->cancel)
+		return BE_CANCELLED;
+	d->nwa = start + buf->sectors;
+	return 0;
+}
+
+
+/* ts A70911 */
+/* Flush stdio system buffer to physical device.
+   @param flag bit0= do not report debug message (intermediate sync)
+*/
+int burn_stdio_sync_cache(int fd, struct burn_drive *d, int flag)
+{
+	if (fd < 0) {
+
+		/* >>> program error */;
+
+		d->cancel = 1;
+		return 0;
+	}
+	d->needs_sync_cache = 0;
+	if (!(flag & 1))
+		libdax_msgs_submit(libdax_messenger, -1, 0x00000002,
+			   LIBDAX_MSGS_SEV_DEBUG, LIBDAX_MSGS_PRIO_ZERO,
+			   "syncing cache (stdio fsync)", 0, 0);
+	if (fsync(fd) != 0) {
+		if (errno == EINVAL) /* E.g. /dev/null cannot fsync */
+			return 1;
+		libdax_msgs_submit(libdax_messenger, d->global_index,
+			0x00020148,
+			LIBDAX_MSGS_SEV_SORRY, LIBDAX_MSGS_PRIO_HIGH,
+			"Cannot write desired amount of data", errno, 0);
+		d->cancel = 1;
+		return 0;
+	}
+	return 1;
+}
+
+
+/* ts A70911 : to be used as burn_drive.sync_cache(),
+               emulating mmc_sync_cache() */
+void burn_stdio_mmc_sync_cache(struct burn_drive *d)
+{
+	burn_stdio_sync_cache(d->stdio_fd, d, 0);
+}
+
+
+/* ts A70912 */
+/* Enforces eventual nominal write speed.
+   @param flag bit0= initialize *prev_time */
+int burn_stdio_slowdown(struct burn_drive *d, struct timeval *prev_time,
+			 int amount, int flag)
+{
+	struct timeval tnow;
+	struct timezone dummy_tz;
+	double to_wait;
+
+	if (flag & 1) {
+		gettimeofday(prev_time, &dummy_tz);
+		return 1;
+	}
+	if(d->nominal_write_speed <= 0)
+		return 2;
+	gettimeofday(&tnow, &dummy_tz);
+	to_wait = ( ((double) amount) / (double) d->nominal_write_speed ) - 
+		(double) ( tnow.tv_sec - prev_time->tv_sec ) -
+		(double) ( tnow.tv_usec - prev_time->tv_usec ) / 1.0e6
+		- 0.001; /* best would be 1 / kernel granularity HZ */
+	if (to_wait >= 0.0001) {
+		usleep((int) (to_wait * 1000000.0));
+	}
+	gettimeofday(prev_time, &dummy_tz);
+	return 1;
+}
+
+
+/* ts A70904 */
+int burn_stdio_write_track(struct burn_write_opts *o, struct burn_session *s,
+				int tnum, int flag)
+{
+	int open_ended, bufsize, ret, sectors, fd;
+	struct burn_track *t = s->track[tnum];
+	struct burn_drive *d = o->drive;
+	char buf[16*2048];
+	int i, prev_sync_sector = 0;
+	struct buffer *out = d->buffer;
+	struct timeval prev_time;
+
+	bufsize = sizeof(buf);
+	fd = d->stdio_fd;
+
+	sectors = burn_track_get_sectors(t);
+	burn_disc_init_track_status(o, s, tnum, sectors);
+	open_ended = burn_track_is_open_ended(t);
+
+	/* attach stdio emulators for mmc_*() functions */
+	if (o->simulate)
+		d->write = burn_stdio_mmc_dummy_write;
+	else
+		d->write = burn_stdio_mmc_write;
+	d->sync_cache = burn_stdio_mmc_sync_cache;
+
+	burn_stdio_slowdown(d, &prev_time, 0, 1); /* initialize */
+	for (i = 0; open_ended || i < sectors; i++) {
+		/* transact a (CD sized) sector */
+		if (!sector_data(o, t, 0))
+			{ret= 0; goto ex;}
+		if (open_ended) {
+			d->progress.sectors = sectors = d->progress.sector;
+			if (burn_track_is_data_done(t))
+	break;
+		}
+		d->progress.sector++;
+		/* Flush to disk after each full MB */
+		if (d->progress.sector - prev_sync_sector >= 512) {
+			prev_sync_sector = d->progress.sector;
+			if (!o->simulate)
+				burn_stdio_sync_cache(d->stdio_fd, d, 1);
+			burn_stdio_slowdown(d, &prev_time, 512 * 2, 0);
+		}
+	}
+
+	/* Pad up buffer to next full o->obs (usually 32 kB) */
+	if (o->obs_pad && out->bytes > 0 && out->bytes < o->obs) {
+		memset(out->data + out->bytes, 0, o->obs - out->bytes);
+		out->sectors += (o->obs - out->bytes) / 2048;
+		out->bytes = o->obs;
+	}
+	ret = burn_write_flush(o, t);
+	ret= 1;
+ex:;
+	if (d->cancel)
+		burn_source_cancel(t->source);
+	return ret;
+}
+
+
+/* ts A70904 */
+int burn_stdio_write_sync(struct burn_write_opts *o,
+				 struct burn_disc *disc)
+{
+	int ret;
+	struct burn_drive *d = o->drive;
+
+	d->needs_close_session = 0;
+	o->obs_pad = 0; /* no filling-up of track's last 32k buffer */
+	o->obs = 32*1024; /* buffer size */
+
+	if (disc->sessions != 1)
+		{ret= 0 ; goto ex;}
+	if (disc->session[0]->tracks != 1)
+		{ret= 0 ; goto ex;}
+	
+	/* update progress */
+	d->progress.session = 0;
+	d->progress.tracks = 1;
+
+	/* >>> adjust sector size (2048) to eventual audio or even raw */
+	/* open target file */
+	if (d->stdio_fd >= 0)
+		close(d->stdio_fd);
+	d->stdio_fd = burn_stdio_open_write(d, o->start_byte, 2048, 0);
+	if (d->stdio_fd == -1)
+		{ret = 0; goto ex;}
+
+	ret = burn_stdio_write_track(o, disc->session[0], 0, 0);
+	if (ret <= 0)
+		goto ex;
+
+	/* XXX: currently signs an end of session */
+	d->progress.sector = 0;
+	d->progress.start_sector = 0;
+	d->progress.sectors = 0;
+	ret = 1;
+ex:;
+	if (d->stdio_fd >= 0)
+		close(d->stdio_fd);
+	d->stdio_fd = -1;
+
+	/* update media state records */
+	burn_drive_mark_unready(d);
+
+	d->busy = BURN_DRIVE_IDLE;
+	return ret;
+}
+
+
 void burn_disc_write_sync(struct burn_write_opts *o, struct burn_disc *disc)
 {
 	struct cue_sheet *sheet;
@@ -1765,6 +1981,13 @@ void burn_disc_write_sync(struct burn_write_opts *o, struct burn_disc *disc)
 	d->rlba = -150;
 	d->toc_temp = 9;
 
+	/* ts A70904 */
+	if (d->drive_role != 1) {
+		ret = burn_stdio_write_sync(o, disc);
+		if (ret <= 0)
+			goto fail_wo_sync;
+		return;
+	}
 	/* ts A61218 */
 	if (! d->current_is_cd_profile) {
 		ret = burn_dvd_write_sync(o, disc);
@@ -1773,45 +1996,12 @@ void burn_disc_write_sync(struct burn_write_opts *o, struct burn_disc *disc)
 		return;
 	}
 
-#ifndef Libburn_precheck_write_ruleS
-	/* <<< covered by burn_precheck_write() */
-	if (o->start_byte >= 0) {
-		sprintf(msg, "Write start address not supported");
-		libdax_msgs_submit(libdax_messenger, d->global_index,
-			0x00020124,
-			LIBDAX_MSGS_SEV_SORRY, LIBDAX_MSGS_PRIO_HIGH,
-			msg, 0, 0);
-		goto fail_wo_sync;
-	}
-#endif /* ! Libburn_precheck_write_ruleS */
-
 	/* ts A70218 */
 	if (o->write_type == BURN_WRITE_SAO) {
-
-#ifndef Libburn_precheck_write_ruleS
-		/* <<< covered by burn_precheck_write() "appended session" */
-		if (disc->sessions > 1) {
-sao_restriction_violated:;
-			libdax_msgs_submit(libdax_messenger, d->global_index,
-				0x0002012f,
-				LIBDAX_MSGS_SEV_SORRY, LIBDAX_MSGS_PRIO_HIGH,
-				"SAO is restricted to a single session with fixed track sizes",
-				0, 0);
-			goto fail_wo_sync;
-		}
-#endif /* ! Libburn_precheck_write_ruleS */
-
 		for (i = 0 ; i < disc->session[0]->tracks; i++) {
 			t = disc->session[0]->track[i];
 			if (burn_track_is_open_ended(t)) {
 				default_size = burn_track_get_default_size(t);
-
-#ifndef Libburn_precheck_write_ruleS
-				/* <<< covered by burn_precheck_write() */
-				if (default_size <= 0)
-					goto sao_restriction_violated;
-#endif /* ! Libburn_precheck_write_ruleS */
-
 				sprintf(msg,
 					"Activated track default size %.f",
 					(double) default_size);
@@ -1969,3 +2159,124 @@ fail_wo_sync:;
 	d->cancel = 1;
 	d->busy = BURN_DRIVE_IDLE;
 }
+
+/* ts A70811 : API function */
+int burn_random_access_write(struct burn_drive *d, off_t byte_address,
+				char *data, off_t data_count, int flag)
+{
+	int alignment = 0, start, upto, chunksize, err, fd = -1, ret;
+	char msg[81], *rpt;
+	struct buffer buf;
+
+	if (d->released) {
+		libdax_msgs_submit(libdax_messenger,
+			d->global_index, 0x00020142,
+			LIBDAX_MSGS_SEV_FATAL, LIBDAX_MSGS_PRIO_HIGH,
+			"Drive is not grabbed on random access write", 0, 0);
+		return 0;
+	}
+	if(d->drive_role == 0) {
+		libdax_msgs_submit(libdax_messenger, d->global_index,
+			0x00020146,
+			LIBDAX_MSGS_SEV_FATAL, LIBDAX_MSGS_PRIO_HIGH,
+			"Drive is a virtual placeholder (null-drive)", 0, 0);
+		return 0;
+	}
+
+	if(d->drive_role == 2)
+		alignment = 2 * 1024;
+	if (d->current_profile == 0x12) /* DVD-RAM */
+		alignment = 2 * 1024;
+        if (d->current_profile == 0x13) /* DVD-RW restricted overwrite */
+		alignment = 32 * 1024;
+	if (d->current_profile == 0x1a) /* DVD+RW */
+		alignment = 2 * 1024;
+	if (alignment == 0) {
+		sprintf(msg, "Write start address not supported");
+		libdax_msgs_submit(libdax_messenger, d->global_index,
+			0x00020125,
+			LIBDAX_MSGS_SEV_SORRY, LIBDAX_MSGS_PRIO_HIGH,
+			"Write start address not supported", 0, 0);
+		return 0;
+	}
+	if ((byte_address % alignment) != 0) {
+		sprintf(msg,
+			"Write start address not properly aligned (%d bytes)",
+			alignment);
+		libdax_msgs_submit(libdax_messenger, d->global_index,
+			0x00020126,
+			LIBDAX_MSGS_SEV_SORRY, LIBDAX_MSGS_PRIO_HIGH,
+			msg, 0, 0);
+		return 0;
+	}
+	if ((data_count % alignment) != 0) {
+		sprintf(msg,
+			"Write data count not properly aligned (%ld bytes)",
+			(long) alignment);
+		libdax_msgs_submit(libdax_messenger, d->global_index,
+			0x00020141,
+			LIBDAX_MSGS_SEV_SORRY, LIBDAX_MSGS_PRIO_HIGH,
+			msg, 0, 0);
+		return 0;
+	}
+	if (d->busy != BURN_DRIVE_IDLE) {
+		libdax_msgs_submit(libdax_messenger,
+			d->global_index, 0x00020140,
+			LIBDAX_MSGS_SEV_FATAL, LIBDAX_MSGS_PRIO_HIGH,
+			"Drive is busy on attempt to write random access",0,0);
+		return 0;
+	}
+	if(d->drive_role != 1) {
+		fd = burn_stdio_open_write(d, byte_address, 2048, 0);
+		if (fd == -1)
+			return 0;
+	}
+	d->busy = BURN_DRIVE_WRITING_SYNC;
+	d->buffer = &buf;
+
+	start = byte_address / 2048;
+	upto = start + data_count / 2048;
+	rpt = data;
+	for (; start < upto; start += 16) {
+		chunksize = upto - start;
+		if (chunksize > 16)
+			chunksize = 16;
+		d->buffer->bytes = chunksize * 2048;
+		memcpy(d->buffer->data, rpt, d->buffer->bytes);
+		rpt += d->buffer->bytes;
+		d->buffer->sectors = chunksize;
+		d->nwa = start;
+		if(d->drive_role == 1) {
+			err = d->write(d, d->nwa, d->buffer);
+		} else {
+			ret = burn_stdio_write(fd, (char *) d->buffer->data,
+						d->buffer->bytes, d, 0);
+			err = 0;
+			if (ret <= 0)
+				err = BE_CANCELLED;
+		}
+		if (err == BE_CANCELLED) {
+			d->busy = BURN_DRIVE_IDLE;
+			if(fd >= 0)
+				close(fd);
+			return (-(start * 2048 - byte_address));
+		}
+	}
+
+	if(d->drive_role == 1)
+		d->needs_sync_cache = 1;
+	if(flag & 1) {
+		if(d->drive_role == 1)
+			d->sync_cache(d);
+		else
+			burn_stdio_sync_cache(fd, d, 0);
+		d->needs_sync_cache = 0;
+	}
+		
+	if(fd >= 0)
+		close(fd);
+	d->buffer = NULL;
+	d->busy = BURN_DRIVE_IDLE;
+	return 1;
+}
+
