@@ -1,13 +1,13 @@
 
 /* test/libburner.c , API illustration of burning data or audio tracks to CD */
-/* Copyright (C) 2005 - 2006 Thomas Schmitt <scdbackup@gmx.net> */
-/* Provided under GPL, see also "License and copyright aspects" at file end */
+/* Copyright (C) 2005 - 2007 Thomas Schmitt <scdbackup@gmx.net> */
+/* Provided under GPLv2,see also "License and copyright aspects" at file end */
 
 
 /**                               Overview 
   
   libburner is a minimal demo application for the library libburn as provided
-  on  http://libburnia.pykix.org . It can list the available devices, can
+  on  http://libburnia-project.org . It can list the available devices, can
   blank a CD-RW or DVD-RW, can format a DVD-RW, can burn to CD-R, CD-RW, DVD-R,
   DVD+R, DVD+RW, DVD-RAM or DVD-RW. Not supported yet: double layer media.
 
@@ -123,21 +123,28 @@ int libburner_aquire_by_adr(char *drive_adr)
 	int ret;
 	char libburn_drive_adr[BURN_DRIVE_ADR_LEN];
 
+	/* Some not-so-harmless drive addresses get blocked in this demo */
+	if (strncmp(drive_adr, "stdio:/dev/fd/", 14) == 0 ||
+		strcmp(drive_adr, "stdio:-") == 0) {
+		fprintf(stderr, "Will not work with pseudo-drive '%s'\n",
+			drive_adr);
+		return 0;
+	}
+
 	/* This tries to resolve links or alternative device files */
 	ret = burn_drive_convert_fs_adr(drive_adr, libburn_drive_adr);	
 	if (ret<=0) {
-		fprintf(stderr,"Address does not lead to a CD burner: '%s'\n",
-			drive_adr);
-		return ret;
+		fprintf(stderr, "Address does not lead to a CD burner: '%s'\n",
+				 drive_adr);
+		return 0;
 	}
-
-	printf("Aquiring drive '%s' ...\n",libburn_drive_adr);
-	ret = burn_drive_scan_and_grab(&drive_list,libburn_drive_adr,1);
+	fprintf(stderr,"Aquiring drive '%s' ...\n", libburn_drive_adr);
+	ret = burn_drive_scan_and_grab(&drive_list, libburn_drive_adr, 1);
 	if (ret <= 0) {
 		fprintf(stderr,"FAILURE with persistent drive address  '%s'\n",
 			libburn_drive_adr);
 	} else {
-		printf("Done\n");
+		fprintf(stderr,"Done\n");
 		drive_is_grabbed = 1;
 	}
 	return ret;
@@ -163,7 +170,7 @@ int libburner_aquire_by_driveno(int *driveno)
 
 	printf("Beginning to scan for devices ...\n");
 	while (!burn_drive_scan(&drive_list, &drive_count))
-		usleep(1002);
+		usleep(100002);
 	if (drive_count <= 0 && *driveno >= 0) {
 		printf("FAILED (no drives found)\n");
 		return 0;
@@ -352,6 +359,10 @@ int libburner_format_row(struct burn_drive *drive)
     To make sure a data image is fully readable on any Linux machine, this
     function adds 300 kiB of padding to the (usualy single) track.
     Audio tracks get padded to complete their last sector.
+    A fifo of 4 MB is installed between each track and its data source.
+    Each of the 4 MB buffers gets allocated automatically as soon as a track
+    begins to be processed and it gets freed as soon as the track is done.
+    The fifos do not wait for buffer fill but writing starts immediately.
 
     In case of external signals expect abort handling of an ongoing burn to
     last up to a minute. Wait the normal burning timespan before any kill -9.
@@ -363,7 +374,7 @@ int libburner_payload(struct burn_drive *drive,
 		      char source_adr[][4096], int source_adr_count,
 		      int multi, int simulate_burn, int all_tracks_type)
 {
-	struct burn_source *data_src;
+	struct burn_source *data_src, *fifo_src[99];
 	struct burn_disc *target_disc;
 	struct burn_session *session;
 	struct burn_write_opts *burn_options;
@@ -372,6 +383,7 @@ int libburner_payload(struct burn_drive *drive,
 	struct burn_progress progress;
 	time_t start_time;
 	int last_sector = 0, padding = 0, trackno, unpredicted_size = 0, fd;
+	int fifo_chunksize = 2352, fifo_chunks = 1783; /* ~ 4 MB fifo */
 	off_t fixed_size;
 	char *adr, reasons[BURN_REASONS_LEN];
 	struct stat stbuf;
@@ -380,6 +392,8 @@ int libburner_payload(struct burn_drive *drive,
 		all_tracks_type = BURN_MODE1;
 		/* a padding of 300 kiB helps to avoid the read-ahead bug */
 		padding = 300*1024;
+		fifo_chunksize = 2048;
+		fifo_chunks = 2048; /* 4 MB fifo */
 	}
 
 	target_disc = burn_disc_create();
@@ -390,6 +404,7 @@ int libburner_payload(struct burn_drive *drive,
 	  tracklist[trackno] = track = burn_track_create();
 	  burn_track_define_data(track, 0, padding, 1, all_tracks_type);
 
+	  /* Open file descriptor to source of track data */
 	  adr = source_adr[trackno];
 	  fixed_size = 0;
 	  if (adr[0] == '-' && adr[1] == 0) {
@@ -403,6 +418,8 @@ int libburner_payload(struct burn_drive *drive,
 	  }
 	  if (fixed_size==0)
 		unpredicted_size = 1;
+
+	  /* Convert this filedescriptor into a burn_source object */
 	  data_src = NULL;
 	  if (fd>=0)
 	  	data_src = burn_fd_source_new(fd, -1, fixed_size);
@@ -414,14 +431,28 @@ int libburner_payload(struct burn_drive *drive,
 				strerror(errno));
 		return 0;
 	  }
-	  if (burn_track_set_source(track, data_src) != BURN_SOURCE_OK) {
+	  /* Install a fifo object on top of that data source object */
+	  fifo_src[trackno] = burn_fifo_source_new(data_src,
+					fifo_chunksize, fifo_chunks, 0);
+	  if (fifo_src[trackno] == NULL) {
+		fprintf(stderr,
+			"FATAL: Could not create fifo object of 4 MB\n");
+		return 0;
+	  }
+
+	  /* Use the fifo object as data source for the track */
+	  if (burn_track_set_source(track, fifo_src[trackno])
+							 != BURN_SOURCE_OK) {
 		printf("FATAL: Cannot attach source object to track object\n");
 		return 0;
 	  }
 
 	  burn_session_add_track(session, track, BURN_POS_END);
 	  printf("Track %d : source is '%s'\n", trackno+1, adr);
+
+	  /* Give up local reference to the data burn_source object */
 	  burn_source_free(data_src);
+	  
         } /* trackno loop end */
 
 	/* Evaluate drive and media */
@@ -461,25 +492,44 @@ int libburner_payload(struct burn_drive *drive,
 
 	burn_write_opts_free(burn_options);
 	while (burn_drive_get_status(drive, NULL) == BURN_DRIVE_SPAWNING)
-		usleep(1002);
+		usleep(100002);
 	while (burn_drive_get_status(drive, &progress) != BURN_DRIVE_IDLE) {
-		if( progress.sectors <= 0 || progress.sector == last_sector)
+		if (progress.sectors <= 0 ||
+		    (progress.sector >= progress.sectors - 1 &&
+	             !unpredicted_size) ||
+		    (unpredicted_size && progress.sector == last_sector))
 			printf(
-			     "Thank you for being patient since %d seconds.\n",
+			     "Thank you for being patient since %d seconds.",
 			     (int) (time(0) - start_time));
 		else if(unpredicted_size)
-			printf("Track %d : sector %d\n", progress.track+1,
+			printf("Track %d : sector %d", progress.track+1,
 				progress.sector);
 		else
-			printf("Track %d : sector %d of %d\n",progress.track+1,
+			printf("Track %d : sector %d of %d",progress.track+1,
 				progress.sector, progress.sectors);
 		last_sector = progress.sector;
+		if (progress.track >= 0 && progress.track < source_adr_count) {
+			int size, free_bytes, ret;
+			char *status_text;
+	
+			ret = burn_fifo_inquire_status(
+				fifo_src[progress.track], &size, &free_bytes,
+				&status_text);
+			if (ret >= 0 ) 
+				printf("  [fifo %s, %2d%% fill]", status_text,
+					(int) (100.0 - 100.0 *
+						((double) free_bytes) /
+						(double) size));
+		} 
+		printf("\n");
 		sleep(1);
 	}
 	printf("\n");
 
-	for (trackno = 0 ; trackno < source_adr_count; trackno++)
+	for (trackno = 0 ; trackno < source_adr_count; trackno++) {
+	  	burn_source_free(fifo_src[trackno]);
 		burn_track_free(tracklist[trackno]);
+	}
 	burn_session_free(session);
 	burn_disc_free(target_disc);
 	if (multi && current_profile != 0x1a && current_profile != 0x13 &&
@@ -620,7 +670,7 @@ int main(int argc, char **argv)
 	if (ret)
 		exit(ret);
 
-	printf("Initializing libburnia.pykix.org ...\n");
+	printf("Initializing libburnia-project.org ...\n");
 	if (burn_initialize())
 		printf("Done\n");
 	else {
